#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the Audyssey settings of Denon AVR receivers.

:copyright: (c) 2020 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import logging
from collections.abc import Hashable
from typing import List, Optional

import attr

from .appcommand import AppCommandCmd, AppCommandCmdParam, AppCommands
from .const import (
    DENON_ATTR_SETATTR,
    DYNAMIC_VOLUME_MAP,
    DYNAMIC_VOLUME_MAP_LABELS_APPCOMMAND,
    DYNAMIC_VOLUME_MAP_LABELS_TELNET,
    MULTI_EQ_MAP,
    MULTI_EQ_MAP_LABELS_APPCOMMAND,
    MULTI_EQ_MAP_LABELS_TELNET,
    REF_LVL_OFFSET_MAP,
    REF_LVL_OFFSET_MAP_LABELS_APPCOMMAND,
    REF_LVL_OFFSET_MAP_LABELS_TELNET,
)
from .exceptions import AvrCommandError, AvrProcessingError
from .foundation import DenonAVRFoundation, convert_string_int_bool

_LOGGER = logging.getLogger(__name__)


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRAudyssey(DenonAVRFoundation):
    """Audyssey Settings."""

    _dynamiceq: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool), default=None
    )
    _dynamiceq_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool), default=None
    )
    _reflevoffset: Optional[str] = attr.ib(
        converter=attr.converters.optional(REF_LVL_OFFSET_MAP.get), default=None
    )
    _reflevoffset_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool), default=None
    )
    _dynamicvol: Optional[str] = attr.ib(
        converter=attr.converters.optional(DYNAMIC_VOLUME_MAP.get), default=None
    )
    _dynamicvol_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool), default=None
    )
    _multeq: Optional[str] = attr.ib(
        converter=attr.converters.optional(MULTI_EQ_MAP.get), default=None
    )
    _multeq_control: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool), default=None
    )
    _lfc: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_string_int_bool), default=None
    )
    _containment_amount: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )

    # Update tags for attributes
    # AppCommand0300.xml interface
    appcommand0300_attrs = {AppCommands.GetAudyssey: None}

    def setup(self) -> None:
        """Ensure that the instance is initialized."""
        # Add tags for a potential AppCommand.xml update
        for tag in self.appcommand0300_attrs:
            self._device.api.add_appcommand0300_update_tag(tag)

        self._device.telnet_api.register_callback(
            "PS", self._async_sound_detail_callback
        )

        self._is_setup = True

    async def _async_sound_detail_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a sound detail change event."""
        if self._device.zone != zone:
            return

        if parameter[0:6] == "REFLEV":
            self._reflevoffset = parameter[7:]
        elif parameter[0:6] == "DYNVOL":
            self._dynamicvol = parameter[7:]
        elif parameter[0:6] == "MULTEQ":
            self._multeq = parameter[7:]
        elif parameter == "DYNEQ ON":
            self._dynamiceq = "1"
        elif parameter == "DYNEQ OFF":
            self._dynamiceq = "0"
        elif parameter == "LFC ON":
            self._lfc = "1"
        elif parameter == "LFC OFF":
            self._lfc = "0"
        elif parameter[:6] == "CNTAMT":
            self._containment_amount = int(parameter[7:])

    async def async_update(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ) -> None:
        """Update Audyssey asynchronously."""
        _LOGGER.debug("Starting Audyssey update")
        # Ensure instance is setup before updating
        if not self._is_setup:
            self.setup()

        # Update state
        await self.async_update_audyssey(global_update=global_update, cache_id=cache_id)
        _LOGGER.debug("Finished Audyssey update")

    async def async_update_audyssey(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ):
        """Update Audyssey status of device."""
        if self._device.use_avr_2016_update is None:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set"
            )

        # Audyssey is only available for avr 2016 update
        if self._device.use_avr_2016_update:
            try:
                await self.async_update_attrs_appcommand(
                    self.appcommand0300_attrs,
                    appcommand0300=True,
                    global_update=global_update,
                    cache_id=cache_id,
                )
            except AvrProcessingError as err:
                # Don't raise an error here, because not all devices support it
                _LOGGER.debug("Updating Audyssey failed: %s", err)

    async def _async_set_audyssey(self, cmd: AppCommandCmd) -> None:
        """Set Audyssey parameter."""
        res = await self._device.api.async_post_appcommand(
            self._device.urls.appcommand0300, (cmd,)
        )

        try:
            if res.find("cmd").text != "OK":
                raise AvrProcessingError(f"SetAudyssey command {cmd.name} failed")
        except AttributeError as err:
            raise AvrProcessingError(f"SetAudyssey command {cmd.name} failed") from err

    ##############
    # Properties #
    ##############
    @property
    def dynamic_eq(self) -> Optional[bool]:
        """Return value of Dynamic EQ."""
        return self._dynamiceq

    @property
    def reference_level_offset(self) -> Optional[str]:
        """Return value of Reference Level Offset."""
        return self._reflevoffset

    @property
    def reference_level_offset_setting_list(self) -> List[str]:
        """Return a list of available reference level offset settings."""
        if self._device.telnet_available:
            return list(REF_LVL_OFFSET_MAP_LABELS_TELNET.keys())
        return list(REF_LVL_OFFSET_MAP_LABELS_APPCOMMAND.keys())

    @property
    def dynamic_volume(self) -> Optional[str]:
        """Return value of Dynamic Volume."""
        return self._dynamicvol

    @property
    def dynamic_volume_setting_list(self) -> List[str]:
        """Return a list of available Dynamic Volume settings."""
        if self._device.telnet_available:
            return list(DYNAMIC_VOLUME_MAP_LABELS_TELNET.keys())
        return list(DYNAMIC_VOLUME_MAP_LABELS_APPCOMMAND.keys())

    @property
    def multi_eq(self) -> Optional[str]:
        """Return value of MultiEQ."""
        return self._multeq

    @property
    def multi_eq_setting_list(self) -> List[str]:
        """Return a list of available MultiEQ settings."""
        if self._device.telnet_available:
            return list(MULTI_EQ_MAP_LABELS_TELNET.keys())
        return list(MULTI_EQ_MAP_LABELS_APPCOMMAND.keys())

    @property
    def lfc(self) -> Optional[bool]:
        """
        Return value of LFC.

        Only available if using Telnet.
        """
        return self._lfc

    @property
    def containment_amount(self) -> Optional[int]:
        """
        Return value of Containment Amount.

        Only available if using Telnet.
        """
        return self._containment_amount

    ##########
    # Setter #
    ##########
    async def async_dynamiceq_off(self) -> None:
        """Turn DynamicEQ off."""
        if self._device.telnet_available:
            telnet_command = self._device.telnet_commands.command_dynamiceq + "OFF"
            await self._device.telnet_api.async_send_commands(telnet_command)
            return
        cmd = attr.evolve(
            AppCommands.SetAudysseyDynamicEQ,
            param_list=(AppCommandCmdParam(name="dynamiceq", text=0),),
        )
        await self._async_set_audyssey(cmd)

    async def async_dynamiceq_on(self) -> None:
        """Turn DynamicEQ on."""
        if self._device.telnet_available:
            telnet_command = self._device.telnet_commands.command_dynamiceq + "ON"
            await self._device.telnet_api.async_send_commands(telnet_command)
            return
        cmd = attr.evolve(
            AppCommands.SetAudysseyDynamicEQ,
            param_list=(AppCommandCmdParam(name="dynamiceq", text=1),),
        )
        await self._async_set_audyssey(cmd)

    async def async_set_multieq(self, value: str) -> None:
        """Set MultiEQ mode."""
        if self._device.telnet_available:
            setting = MULTI_EQ_MAP_LABELS_TELNET.get(value)
            if setting is None:
                raise AvrCommandError(f"Value {value} not known for MultiEQ")
            telnet_command = self._device.telnet_commands.command_multieq + setting
            await self._device.telnet_api.async_send_commands(telnet_command)
            return

        setting = MULTI_EQ_MAP_LABELS_APPCOMMAND.get(value)
        if setting is None:
            raise AvrCommandError(f"Value {value} not known for MultiEQ")
        cmd = attr.evolve(
            AppCommands.SetAudysseyMultiEQ,
            param_list=(AppCommandCmdParam(name="multeq", text=setting),),
        )
        await self._async_set_audyssey(cmd)

    async def async_set_reflevoffset(self, value: str) -> None:
        """Set Reference Level Offset."""
        # Reference level offset can only be used with DynamicEQ
        if not self._dynamiceq:
            raise AvrCommandError(
                "Reference level could only be set when DynamicEQ is active"
            )
        if self._device.telnet_available:
            setting = REF_LVL_OFFSET_MAP_LABELS_TELNET.get(value)
            if setting is None:
                raise AvrCommandError(
                    f"Value {value} not known for Reference level offset"
                )
            telnet_command = self._device.telnet_commands.command_reflevoffset + setting
            await self._device.telnet_api.async_send_commands(telnet_command)
            return

        setting = REF_LVL_OFFSET_MAP_LABELS_APPCOMMAND.get(value)
        if setting is None:
            raise AvrCommandError(f"Value {value} not known for Reference level offset")
        cmd = attr.evolve(
            AppCommands.SetAudysseyReflevoffset,
            param_list=(AppCommandCmdParam(name="reflevoffset", text=setting),),
        )
        await self._async_set_audyssey(cmd)

    async def async_set_dynamicvol(self, value: str) -> None:
        """Set Dynamic Volume."""
        if self._device.telnet_available:
            setting = DYNAMIC_VOLUME_MAP_LABELS_TELNET.get(value)
            if setting is None:
                raise AvrCommandError(f"Value {value} not known for Dynamic Volume")
            telnet_command = self._device.telnet_commands.command_dynamicvol + setting
            await self._device.telnet_api.async_send_commands(telnet_command)
            return

        setting = DYNAMIC_VOLUME_MAP_LABELS_APPCOMMAND.get(value)
        if setting is None:
            raise AvrCommandError(f"Value {value} not known for Dynamic Volume")
        cmd = attr.evolve(
            AppCommands.SetAudysseyDynamicvol,
            param_list=(AppCommandCmdParam(name="dynamicvol", text=setting),),
        )
        await self._async_set_audyssey(cmd)

    async def async_toggle_dynamic_eq(self) -> None:
        """Toggle DynamicEQ."""
        if self._dynamiceq:
            await self.async_dynamiceq_off()
        else:
            await self.async_dynamiceq_on()

    async def async_lfc_on(self):
        """Turn LFC on."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_lfc.format(mode="ON")
            )
            return
        await self._device.api.async_get_command(
            self._device.urls.command_lfc.format(mode="ON")
        )

    async def async_lfc_off(self):
        """Turn LFC off."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_lfc.format(mode="OFF")
            )
            return
        await self._device.api.async_get_command(
            self._device.urls.command_lfc.format(mode="OFF")
        )

    async def async_toggle_lfc(self):
        """Toggle LFC."""
        if self._lfc:
            await self.async_lfc_off()
        else:
            await self.async_lfc_on()

    async def async_containment_amount(self, amount: int) -> None:
        """
        Set Containment Amount.

        Valid values are 1-7.
        """
        if amount < 1 or amount > 7:
            raise AvrCommandError("Containment amount must be between 1 and 7")
        local_amount = f"{amount:02}"
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_containment_amount.format(
                    value=local_amount
                )
            )
            return
        await self._device.api.async_get_command(
            self._device.urls.command_containment_amount.format(value=local_amount)
        )

    async def async_containment_amount_up(self) -> None:
        """Increase Containment Amount."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_containment_amount.format(
                    value="UP"
                )
            )
            return
        await self._device.api.async_get_command(
            self._device.urls.command_containment_amount.format(value="UP")
        )

    async def async_containment_amount_down(self) -> None:
        """Decrease Containment Amount."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_containment_amount.format(
                    value="DOWN"
                )
            )
            return
        await self._device.api.async_get_command(
            self._device.urls.command_containment_amount.format(value="DOWN")
        )


def audyssey_factory(instance: DenonAVRFoundation) -> DenonAVRAudyssey:
    """Create  DenonAVRAudyssey at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRAudyssey(device=instance._device)
    return new



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the foundation classes for Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import asyncio
import logging
import xml.etree.ElementTree as ET
from collections.abc import Hashable
from copy import deepcopy
from typing import Dict, List, Literal, Optional, Union, get_args

import attr

from .api import DenonAVRApi, DenonAVRTelnetApi
from .appcommand import AppCommandCmd, AppCommands
from .const import (
    APPCOMMAND_CMD_TEXT,
    APPCOMMAND_NAME,
    AUDIO_RESTORER_MAP,
    AUDIO_RESTORER_MAP_LABELS,
    AVR,
    AVR_X,
    AVR_X_2016,
    BLUETOOTH_OUTPUT_MAP_LABELS,
    BLUETOOTH_OUTPUT_MODES_MAP,
    CHANNEL_VOLUME_MAP,
    DENON_ATTR_SETATTR,
    DENONAVR_TELNET_COMMANDS,
    DENONAVR_URLS,
    DESCRIPTION_TYPES,
    DEVICEINFO_AVR_X_PATTERN,
    DEVICEINFO_COMMAPI_PATTERN,
    DIMMER_MODE_MAP,
    DIMMER_MODE_MAP_LABELS,
    ECO_MODE_MAP,
    ECO_MODE_MAP_LABELS,
    HDMI_OUTPUT_MAP,
    HDMI_OUTPUT_MAP_LABELS,
    MAIN_ZONE,
    POWER_STATES,
    SETTINGS_MENU_STATES,
    VALID_RECEIVER_TYPES,
    VALID_ZONES,
    VIDEO_PROCESSING_MODES_MAP,
    VIDEO_PROCESSING_MODES_MAP_LABELS,
    ZONE2,
    ZONE2_TELNET_COMMANDS,
    ZONE2_URLS,
    ZONE3,
    ZONE3_TELNET_COMMANDS,
    ZONE3_URLS,
    AudioRestorers,
    AutoStandbys,
    BluetoothOutputModes,
    DimmerModes,
    EcoModes,
    HDMIAudioDecodes,
    HDMIOutputs,
    PanelLocks,
    ReceiverType,
    ReceiverURLs,
    RoomSizes,
    TelnetCommands,
    TransducerLPFs,
    VideoProcessingModes,
)
from .exceptions import (
    AvrCommandError,
    AvrForbiddenError,
    AvrIncompleteResponseError,
    AvrNetworkError,
    AvrProcessingError,
    AvrRequestError,
    AvrTimoutError,
)
from .ssdp import evaluate_scpd_xml

_LOGGER = logging.getLogger(__name__)


def convert_on_off_bool(value: str) -> Optional[bool]:
    """Convert a ON/OFF string to bool."""
    if value is None:
        return None
    if value.lower() == "on":
        return True
    if value.lower() == "off":
        return False
    return None


def convert_on_off_bool_str(value: str) -> Optional[Union[bool, str]]:
    """Convert a ON/OFF string to bool with fallback to raw value."""
    val = convert_on_off_bool(value)
    if val is not None:
        return val

    return value


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRDeviceInfo:
    """Implements a class with device information of the receiver."""

    api: DenonAVRApi = attr.ib(
        validator=attr.validators.instance_of(DenonAVRApi),
        default=attr.Factory(DenonAVRApi),
        kw_only=True,
    )
    telnet_api: DenonAVRTelnetApi = attr.ib(
        validator=attr.validators.instance_of(DenonAVRTelnetApi),
        default=attr.Factory(DenonAVRTelnetApi),
        kw_only=True,
    )
    receiver: Optional[ReceiverType] = attr.ib(
        validator=attr.validators.optional(attr.validators.in_(VALID_RECEIVER_TYPES)),
        default=None,
    )
    telnet_commands: TelnetCommands = attr.ib(
        validator=attr.validators.instance_of(TelnetCommands), init=False
    )
    urls: ReceiverURLs = attr.ib(
        validator=attr.validators.instance_of(ReceiverURLs), init=False
    )
    zone: str = attr.ib(
        validator=attr.validators.in_(VALID_ZONES), default=MAIN_ZONE, kw_only=True
    )
    friendly_name: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    manufacturer: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    model_name: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    serial_number: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    use_avr_2016_update: Optional[bool] = attr.ib(
        converter=attr.converters.optional(bool), default=None
    )
    _power: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _settings_menu: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _dimmer: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _dimmer_modes = get_args(DimmerModes)
    _auto_standby: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _auto_standbys = get_args(AutoStandbys)
    _sleep: Optional[Union[str, int]] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _delay: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    _eco_mode: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _eco_modes = get_args(EcoModes)
    _hdmi_output: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _hdmi_outputs = get_args(HDMIOutputs)
    _hdmi_audio_decode: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _hdmi_audio_decodes = get_args(HDMIAudioDecodes)
    _video_processing_mode: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _video_processing_modes = get_args(VideoProcessingModes)
    _tactile_transducer: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _tactile_transducer_level: Optional[float] = attr.ib(
        converter=attr.converters.optional(float), default=None
    )
    _tactile_transducer_lpf: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _tactile_transducer_lpfs = get_args(TransducerLPFs)
    _room_size: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _room_sizes = get_args(RoomSizes)
    _triggers: Optional[Dict[int, str]] = attr.ib(default=None)
    _speaker_preset: Optional[int] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _bt_transmitter: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _bt_output_mode: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _bt_output_modes = get_args(BluetoothOutputModes)
    _delay_time: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    _audio_restorer: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _audio_restorers = get_args(AudioRestorers)
    _panel_locks = get_args(PanelLocks)
    _graphic_eq: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _headphone_eq: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _is_setup: bool = attr.ib(converter=bool, default=False, init=False)
    _allow_recovery: bool = attr.ib(converter=bool, default=True, init=True)
    _setup_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))

    def __attrs_post_init__(self) -> None:
        """Initialize special attributes and callbacks."""
        # URLs depending from value of self.zone attribute
        if self.zone == MAIN_ZONE:
            self.telnet_commands = DENONAVR_TELNET_COMMANDS
            self.urls = DENONAVR_URLS
        elif self.zone == ZONE2:
            self.telnet_commands = ZONE2_TELNET_COMMANDS
            self.urls = ZONE2_URLS
        elif self.zone == ZONE3:
            self.telnet_commands = ZONE3_TELNET_COMMANDS
            self.urls = ZONE3_URLS
        else:
            raise ValueError(f"Invalid zone {self.zone}")

    async def _async_power_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a power change event."""
        if self.zone == zone and parameter in POWER_STATES:
            self._power = parameter

    async def _async_settings_menu_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a settings menu event."""
        if (
            event == "MN"
            and parameter[0:3] == "MEN"
            and parameter[4:] in SETTINGS_MENU_STATES
        ):
            self._settings_menu = parameter[4:]

    async def _async_dimmer_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a dimmer change event."""
        if event == "DIM" and parameter[1:] in DIMMER_MODE_MAP_LABELS:
            self._dimmer = DIMMER_MODE_MAP_LABELS[parameter[1:]]

    async def _async_auto_standby_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a auto standby change event."""
        if zone == "Main" and event == "STBY":
            self._auto_standby = parameter

    async def _async_auto_sleep_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a sleep change event."""
        if event != "SLP":
            return

        if parameter == "OFF":
            self._sleep = parameter
        else:
            self._sleep = int(parameter)

    async def _async_room_size_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a room size change event."""
        if parameter[:3] != "RSZ":
            return

        self._room_size = parameter[4:]

    async def _async_trigger_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a trigger change event."""
        if event != "TR":
            return

        values = parameter.split()
        if len(values) != 2:
            return

        if self._triggers is None:
            self._triggers = {}

        self._triggers[int(values[0])] = values[1]

    async def _async_delay_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a delay change event."""
        if event == "PS" and parameter[0:5] == "DELAY":
            self._delay = int(parameter[6:])

    async def _async_eco_mode_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle an Eco-mode change event."""
        if event == "ECO" and parameter in ECO_MODE_MAP_LABELS:
            self._eco_mode = ECO_MODE_MAP_LABELS[parameter]

    async def _async_hdmi_output_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a HDMI output change event."""
        if event == "VS" and parameter[0:4] == "MONI":
            self._hdmi_output = HDMI_OUTPUT_MAP_LABELS[parameter]

    async def _async_hdmi_audio_decode_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a HDMI Audio Decode mode change event."""
        if event == "VS" and parameter[0:5] == "AUDIO":
            self._hdmi_audio_decode = parameter[6:]

    async def _async_video_processing_mode_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Video Processing Mode change event."""
        if event == "VS" and parameter[0:3] == "VPM":
            self._video_processing_mode = VIDEO_PROCESSING_MODES_MAP_LABELS[
                parameter[3:]
            ]

    async def _async_tactile_transducer_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a tactile transducer change event."""
        key_value = parameter.split()
        if len(key_value) != 2 or parameter[0:3] != "TTR":
            return

        key = key_value[0]
        value = key_value[1]
        if value == "END":
            return

        if key == "TTR":
            self._tactile_transducer = value
        elif key == "TTRLEV":
            self._tactile_transducer_level = CHANNEL_VOLUME_MAP[value]
        elif key == "TTRLPF":
            self._tactile_transducer_lpf = f"{int(value)} Hz"

    async def _async_speaker_preset_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a speaker preset change event."""
        if event != "SP":
            return

        if parameter[0:2] == "PR":
            self._speaker_preset = int(parameter[3:])

    async def _async_bt_callback(self, zone: str, event: str, parameter: str) -> None:
        """Handle a Bluetooth change event."""
        if event != "BT" or parameter[0:2] != "TX":
            return

        if parameter[3:] in ("ON", "OFF"):
            self._bt_transmitter = parameter[3:]
        else:
            self._bt_output_mode = BLUETOOTH_OUTPUT_MAP_LABELS[parameter[3:]]

    async def _async_delay_time_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a delay time change event."""
        # do not match "DELAY" as it's another event
        if event != "PS" or parameter[0:3] != "DEL" or parameter[0:5] == "DELAY":
            return

        self._delay_time = int(parameter[4:])

    async def _async_audio_restorer_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle an audio restorer change event."""
        if event != "PS" or parameter[0:4] != "RSTR":
            return

        self._audio_restorer = AUDIO_RESTORER_MAP_LABELS[parameter[5:]]

    async def _async_graphic_eq_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Graphic EQ change event."""
        if parameter[0:3] != "GEQ":
            return

        self._graphic_eq = parameter[4:]

    async def _async_headphone_eq_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Headphone EQ change event."""
        if parameter[0:3] != "HEQ":
            return

        self._headphone_eq = parameter[4:]

    def get_own_zone(self) -> str:
        """
        Get zone from actual instance.

        These zone information are used to evaluate responses of HTTP POST
        commands.
        """
        if self.zone == MAIN_ZONE:
            return "zone1"
        return self.zone.lower()

    async def async_setup(self) -> None:
        """Ensure that configuration is loaded from receiver asynchronously."""
        async with self._setup_lock:
            _LOGGER.debug("Starting device setup")
            # Reduce read timeout during receiver identification
            # deviceinfo endpoint takes very long to return 404
            read_timeout = self.api.read_timeout
            self.api.read_timeout = self.api.timeout
            try:
                _LOGGER.debug("Identifying receiver")
                await self.async_identify_receiver()
                _LOGGER.debug("Getting device info")
                await self.async_get_device_info()
            finally:
                self.api.read_timeout = read_timeout
            _LOGGER.debug("Identifying update method")
            await self.async_identify_update_method()

            # Add tags for a potential AppCommand.xml update
            self.api.add_appcommand_update_tag(AppCommands.GetAllZonePowerStatus)

            power_event = "ZM"
            if self.zone == ZONE2:
                power_event = "Z2"
            elif self.zone == ZONE3:
                power_event = "Z3"
            self.telnet_api.register_callback(power_event, self._async_power_callback)

            self.telnet_api.register_callback("MN", self._async_settings_menu_callback)
            self.telnet_api.register_callback("DIM", self._async_dimmer_callback)
            self.telnet_api.register_callback("PS", self._async_delay_callback)
            self.telnet_api.register_callback("ECO", self._async_eco_mode_callback)
            self.telnet_api.register_callback("VS", self._async_hdmi_output_callback)
            self.telnet_api.register_callback(
                "VS", self._async_hdmi_audio_decode_callback
            )
            self.telnet_api.register_callback(
                "VS", self._async_video_processing_mode_callback
            )
            self.telnet_api.register_callback(
                "SS", self._async_tactile_transducer_callback
            )
            self.telnet_api.register_callback("STBY", self._async_auto_standby_callback)
            self.telnet_api.register_callback("SLP", self._async_auto_sleep_callback)
            self.telnet_api.register_callback("PS", self._async_room_size_callback)
            self.telnet_api.register_callback("TR", self._async_trigger_callback)
            self.telnet_api.register_callback("SP", self._async_speaker_preset_callback)
            self.telnet_api.register_callback("BT", self._async_bt_callback)
            self.telnet_api.register_callback("PS", self._async_delay_time_callback)
            self.telnet_api.register_callback("PS", self._async_audio_restorer_callback)
            self.telnet_api.register_callback("PS", self._async_graphic_eq_callback)
            self.telnet_api.register_callback("PS", self._async_headphone_eq_callback)

            self._is_setup = True
            _LOGGER.debug("Finished device setup")

    async def async_update(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ) -> None:
        """Update status asynchronously."""
        _LOGGER.debug("Starting device update")
        # Ensure instance is setup before updating
        if not self._is_setup:
            await self.async_setup()

        # Update power status
        await self.async_update_power(global_update=global_update, cache_id=cache_id)
        _LOGGER.debug("Finished device update")

    async def async_identify_receiver(self) -> None:
        """Identify receiver asynchronously."""
        # Test Deviceinfo.xml if receiver is an AVR-X with port 80 for pre 2016
        # devices and port 8080 devices 2016 and later
        # 2016 models has also some of the XML but not all, try first 2016
        r_types = [AVR_X, AVR_X_2016]

        timeout_errors = 0
        for r_type in r_types:
            self.api.port = r_type.port
            # This XML is needed to get the sources of the receiver
            try:
                # Deviceinfo.xml is static and can be cached for the whole time
                xml = await self.api.async_get_xml(
                    self.urls.deviceinfo, cache_id=id(self)
                )
            except (AvrTimoutError, AvrNetworkError) as err:
                _LOGGER.debug(
                    "Connection error on port %s when identifying receiver: %s",
                    r_type.port,
                    err,
                )

                # Raise error only when occurred at both types
                timeout_errors += 1
                if timeout_errors == len(r_types):
                    raise

            except AvrRequestError as err:
                _LOGGER.debug(
                    (
                        "Request error on port %s when identifying receiver, "
                        "device is not a %s receiver: %s"
                    ),
                    r_type.port,
                    r_type.type,
                    err,
                )
            else:
                is_avr_x = self._is_avr_x(xml)
                if is_avr_x:
                    self.receiver = r_type
                    _LOGGER.info(
                        "Identified %s receiver using port %s",
                        r_type.type,
                        r_type.port,
                    )
                    # Receiver identified, return
                    return

        # If check of Deviceinfo.xml was not successful, receiver is type AVR
        self.receiver = AVR
        self.api.port = AVR.port
        _LOGGER.info(
            "Identified %s receiver using port %s",
            AVR.type,
            AVR.port,
        )

    @staticmethod
    def _is_avr_x(deviceinfo: ET.Element) -> bool:
        """Evaluate Deviceinfo.xml if the device is an AVR-X device."""
        # First test by CommApiVers
        try:
            if bool(
                DEVICEINFO_COMMAPI_PATTERN.search(deviceinfo.find("CommApiVers").text)
                is not None
            ):
                # receiver found , return True
                return True
        except AttributeError:
            # AttributeError occurs when ModelName tag is not found.
            # In this case there is no AVR-X device
            pass

        # if first test did not find AVR-X device, check by model name
        try:
            if bool(
                DEVICEINFO_AVR_X_PATTERN.search(deviceinfo.find("ModelName").text)
                is not None
            ):
                # receiver found , return True
                return True
        except AttributeError:
            # AttributeError occurs when ModelName tag is not found
            # In this case there is no AVR-X device
            pass

        return False

    async def async_identify_update_method(self) -> None:
        """
        Identify the correct update method for the receiver asynchronously.

        As a result this method sets friendly name too.
        """
        # AVR receivers do not support AppCommand.xml interface
        if self.receiver == AVR:
            self.use_avr_2016_update = False
            _LOGGER.info("AVR device, AppCommand.xml interface not supported")
        else:
            try:
                xml = await self.api.async_post_appcommand(
                    self.urls.appcommand, (AppCommands.GetFriendlyName,)
                )
            except (AvrTimoutError, AvrNetworkError) as err:
                _LOGGER.debug(
                    "Connection error when identifying update method: %s", err
                )
                raise
            except AvrRequestError as err:
                _LOGGER.debug("Request error when identifying update method: %s", err)
                self.use_avr_2016_update = False
                _LOGGER.info("AVR-X device, AppCommand.xml interface not supported")
            else:
                self.use_avr_2016_update = True
                _LOGGER.info("AVR-X device, using AppCommand.xml interface")
                self._set_friendly_name(xml)

        if not self.use_avr_2016_update:
            try:
                xml = await self.api.async_get_xml(self.urls.mainzone)
            except (AvrTimoutError, AvrNetworkError) as err:
                _LOGGER.debug(
                    "Connection error when identifying update method: %s", err
                )
                raise
            except AvrRequestError as err:
                _LOGGER.debug("Request error getting friendly name: %s", err)
                _LOGGER.info(
                    "Receiver name could not be determined. Using standard"
                    " name: Denon AVR"
                )
                if self.friendly_name is None:
                    self.friendly_name = "Denon AVR"
            else:
                self._set_friendly_name(xml)

    async def async_verify_avr_2016_update_method(
        self, *, cache_id: Hashable = None
    ) -> None:
        """Verify if avr 2016 update method is working."""
        # Nothing to do if Appcommand.xml interface is not supported
        if self._is_setup and not self.use_avr_2016_update:
            return

        try:
            # Result is cached that it can be reused during update
            await self.api.async_get_global_appcommand(cache_id=cache_id)
        except (AvrTimoutError, AvrNetworkError) as err:
            _LOGGER.debug("Connection error when verifying update method: %s", err)
            raise
        except AvrForbiddenError:
            # Recovery in case receiver changes port from 80 to 8080 which
            # might happen at Denon AVR-X 2016 receivers
            if self._allow_recovery:
                self._allow_recovery = False
                _LOGGER.warning(
                    "AppCommand.xml returns HTTP status 403. Running setup"
                    " again once to check if receiver interface switched "
                    "ports"
                )
                self._is_setup = False
                await self.async_setup()
                await self.async_verify_avr_2016_update_method(cache_id=cache_id)
            else:
                raise
        except AvrIncompleteResponseError as err:
            _LOGGER.debug("Request error when verifying update method: %s", err)
            # Only AVR_X devices support both interfaces
            if self.receiver == AVR_X:
                _LOGGER.warning(
                    "Error verifying Appcommand.xml update method, it returns "
                    "an incomplete result set. Deactivating the interface"
                )
                self.use_avr_2016_update = False
        else:
            if not self._allow_recovery:
                _LOGGER.info("AppCommand.xml recovered from HTTP status 403 error")
            self._allow_recovery = True

    def _set_friendly_name(self, xml: ET.Element) -> None:
        """Set FriendlyName from result xml."""
        # friendlyname tag of AppCommand.xml, FriendlyName tag main zone xml
        tags = ("./cmd/friendlyname", "./FriendlyName/value")
        for tag in tags:
            try:
                name = xml.find(tag).text
            except AttributeError:
                pass
            else:
                if name is not None:
                    self.friendly_name = name.strip()
                    break
        if self.friendly_name is None:
            _LOGGER.warning("No FriendlyName found, using standard name: Denon AVR")
            self.friendly_name = "Denon AVR"

    async def async_get_device_info(self) -> None:
        """Get device information."""
        port = DESCRIPTION_TYPES[self.receiver.type].port
        command = DESCRIPTION_TYPES[self.receiver.type].url
        url = f"http://{self.api.host}:{port}{command}"

        device_info = None
        try:
            res = await self.api.async_get(command, port=port)
        except AvrTimoutError as err:
            _LOGGER.debug("Timeout when getting device info: %s", err)
            raise
        except AvrNetworkError as err:
            _LOGGER.debug("Network error getting device info: %s", err)
            raise
        except AvrRequestError as err:
            _LOGGER.error(
                (
                    "During DenonAVR device identification, when trying to request"
                    " %s the following error occurred: %s"
                ),
                url,
                err,
            )
        else:
            device_info = evaluate_scpd_xml(url, res.text)

        if device_info is None:
            self.manufacturer = "Denon"
            self.model_name = "Unknown"
            self.serial_number = None
            _LOGGER.warning(
                (
                    "Unable to get device information of host %s, Device might be "
                    "in a corrupted state. Continuing without device information. "
                    "Disconnect and reconnect power to the device and try again."
                ),
                self.api.host,
            )
            return

        if self.friendly_name is None and "friendlyName" in device_info:
            self.friendly_name = device_info["friendlyName"]
        self.manufacturer = device_info["manufacturer"]
        self.model_name = device_info["modelName"]
        self.serial_number = device_info["serialNumber"]

    async def async_update_power(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ) -> None:
        """Update power status of device."""
        if self.use_avr_2016_update is None:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set"
            )

        if self.use_avr_2016_update:
            await self.async_update_power_appcommand(
                global_update=global_update, cache_id=cache_id
            )
        else:
            await self.async_update_power_status_xml(cache_id=cache_id)

    async def async_update_power_appcommand(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ) -> None:
        """Update power status from AppCommand.xml."""
        power_appcommand = AppCommands.GetAllZonePowerStatus
        try:
            if global_update:
                xml = await self.api.async_get_global_appcommand(cache_id=cache_id)
            else:
                xml = await self.api.async_post_appcommand(
                    self.urls.appcommand, tuple(power_appcommand), cache_id=cache_id
                )
        except AvrRequestError as err:
            _LOGGER.debug("Error when getting power status: %s", err)
            raise

        # Extract relevant information
        zone = self.get_own_zone()

        # Search for power tag
        power_tag = xml.find(
            f"./cmd[@{APPCOMMAND_CMD_TEXT}='{power_appcommand.cmd_text}']/{zone}"
        )

        if power_tag is None:
            raise AvrProcessingError(
                f"Power attribute of zone {self.zone} not found on update"
            )

        self._power = power_tag.text

    async def async_update_power_status_xml(
        self, cache_id: Optional[Hashable] = None
    ) -> None:
        """Update power status from status xml."""
        # URLs to be scanned
        urls = [self.urls.status]
        if self.zone == MAIN_ZONE:
            urls.append(self.urls.mainzone)
        else:
            urls.append(f"{self.urls.mainzone}?ZoneName={self.zone}")
        # Tags in XML which might contain information about zones power status
        # ordered by their priority
        tags = ["./ZonePower/value", "./Power/value"]

        for tag in tags:
            for url in urls:
                try:
                    xml = await self.api.async_get_xml(url, cache_id=cache_id)
                except AvrRequestError as err:
                    _LOGGER.debug(
                        "Error when getting power status from url %s: %s", url, err
                    )
                    continue

                # Search for power tag
                power_tag = xml.find(tag)
                if power_tag is not None and power_tag.text is not None:
                    self._power = power_tag.text
                    return

        raise AvrProcessingError(
            f"Power attribute of zone {self.zone} not found on update"
        )

    ##############
    # Properties #
    ##############
    @property
    def power(self) -> Optional[str]:
        """
        Return the power state of the device.

        Possible values are: "ON", "STANDBY" and "OFF"
        """
        return self._power

    @property
    def settings_menu(self) -> Optional[bool]:
        """
        Returns the settings menu state of the device.

        Only available if using Telnet.
        """
        return self._settings_menu

    @property
    def dimmer(self) -> Optional[str]:
        """
        Returns the dimmer state of the device.

        Only available if using Telnet.

        Possible values are: "Off", "Dark", "Dim" and "Bright"
        """
        return self._dimmer

    @property
    def auto_standby(self) -> Optional[str]:
        """
        Return the auto-standby state of the device.

        Only available if using Telnet.

        Possible values are: "OFF", "15M", "30M", "60M"
        """
        return self._auto_standby

    @property
    def sleep(self) -> Optional[Union[str, int]]:
        """
        Return the sleep timer for the device.

        Only available if using Telnet.

        Possible values are: "OFF" and 1-120 (in minutes)
        """
        return self._sleep

    @property
    def delay(self) -> Optional[int]:
        """
        Return the audio delay for the device in ms.

        Only available if using Telnet.
        """
        return self._delay

    @property
    def eco_mode(self) -> Optional[str]:
        """
        Returns the eco-mode for the device.

        Only available if using Telnet.

        Possible values are: "Off", "On", "Auto"
        """
        return self._eco_mode

    @property
    def hdmi_output(self) -> Optional[str]:
        """
        Returns the HDMI-output for the device.

        Only available if using Telnet.

        Possible values are: "Auto", "HDMI1", "HDMI2"
        """
        return self._hdmi_output

    @property
    def hdmi_audio_decode(self) -> Optional[str]:
        """
        Returns the HDMI Audio Decode for the device.

        Only available if using Telnet.

        Possible values are: "AMP", "TV"
        """
        return self._hdmi_audio_decode

    @property
    def video_processing_mode(self) -> Optional[str]:
        """
        Return the video processing mode for the device.

        Only available if using Telnet.

        Possible values are: "Auto", "Game", "Movie", "Bypass"
        """
        return self._video_processing_mode

    @property
    def tactile_transducer(self) -> Optional[str]:
        """
        Return the tactile transducer state of the device.

        Only available if using Telnet.
        """
        return self._tactile_transducer

    @property
    def tactile_transducer_level(self) -> Optional[float]:
        """
        Return the tactile transducer level in dB.

        Only available if using Telnet.
        """
        return self._tactile_transducer_level

    @property
    def tactile_transducer_lpf(self) -> Optional[str]:
        """
        Return the tactile transducer low pass filter frequency.

        Only available if using Telnet.
        """
        return self._tactile_transducer_lpf

    @property
    def room_size(self) -> Optional[str]:
        """
        Return the room size for the device.

        Only available if using Telnet.

        Possible values are: "S", "MS", "M", "ML", "L"
        """
        return self._room_size

    @property
    def triggers(self) -> Dict[int, str]:
        """
        Return the triggers and their statuses for the device.

        Only available if using Telnet.
        """
        return self._triggers

    @property
    def speaker_preset(self) -> Optional[int]:
        """
        Return the speaker preset for the device.

        Only available if using Telnet.

        Possible values are: "1", "2"
        """
        return self._speaker_preset

    @property
    def bt_transmitter(self) -> Optional[bool]:
        """
        Return the Bluetooth transmitter state for the device.

        Only available if using Telnet.
        """
        return self._bt_transmitter

    @property
    def bt_output_mode(self) -> Optional[str]:
        """
        Return the Bluetooth output mode for the device.

        Only available if using Telnet.

        Possible values are: "Bluetooth + Speakers", "Bluetooth Only"
        """
        return self._bt_output_mode

    @property
    def delay_time(self) -> Optional[int]:
        """
        Return the delay time for the device in ms.

        Only available if using Telnet.
        """
        return self._delay_time

    @property
    def audio_restorer(self) -> Optional[str]:
        """
        Return the audio restorer for the device.

        Only available if using Telnet.

        Possible values are: "Off", "Low", "Medium", "High"
        """
        return self._audio_restorer

    @property
    def graphic_eq(self) -> Optional[bool]:
        """
        Return the Graphic EQ status for the device.

        Only available if using Telnet.
        """
        return self._graphic_eq

    @property
    def headphone_eq(self) -> Optional[bool]:
        """
        Return the Headphone EQ status for the device.

        Only available if using Telnet.
        """
        return self._headphone_eq

    @property
    def telnet_available(self) -> bool:
        """Return true if telnet is connected and healthy."""
        return self.telnet_api.connected and self.telnet_api.healthy

    ##########
    # Getter #
    ##########

    def get_trigger(self, trigger: int) -> Optional[str]:
        """
        Return the status of a specific trigger.

        Only available if using Telnet.

        Valid trigger values are 1-3.
        """
        if trigger < 1 or trigger > 3:
            raise AvrCommandError(f"Invalid trigger {trigger}, must be between 1 and 3")

        if self._triggers is None:
            return None
        return self._triggers.get(trigger)

    ##########
    # Setter #
    ##########

    async def async_power_on(self) -> None:
        """Turn on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_power_on
            )
        else:
            await self.api.async_get_command(self.urls.command_power_on)

    async def async_power_off(self) -> None:
        """Turn off receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_power_standby
            )
        else:
            await self.api.async_get_command(self.urls.command_power_standby)

    async def async_cursor_up(self) -> None:
        """Cursor Up on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_cusor_up, skip_confirmation=True
            )
        else:
            await self.api.async_get_command(self.urls.command_cusor_up)

    async def async_cursor_down(self) -> None:
        """Cursor Down on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_cusor_down, skip_confirmation=True
            )
        else:
            await self.api.async_get_command(self.urls.command_cusor_down)

    async def async_cursor_left(self) -> None:
        """Cursor Left on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_cusor_left, skip_confirmation=True
            )
        else:
            await self.api.async_get_command(self.urls.command_cusor_left)

    async def async_cursor_right(self) -> None:
        """Cursor Right on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_cusor_right, skip_confirmation=True
            )
        else:
            await self.api.async_get_command(self.urls.command_cusor_right)

    async def async_cursor_enter(self) -> None:
        """Cursor Enter on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_cusor_enter, skip_confirmation=True
            )
        else:
            await self.api.async_get_command(self.urls.command_cusor_enter)

    async def async_back(self) -> None:
        """Back command on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_back, skip_confirmation=True
            )
        else:
            await self.api.async_get_command(self.urls.command_back)

    async def async_info(self) -> None:
        """Info OSD on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_info, skip_confirmation=True
            )
        else:
            await self.api.async_get_command(self.urls.command_info)

    async def async_options(self) -> None:
        """Options menu on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_options, skip_confirmation=True
            )
        else:
            await self.api.async_get_command(self.urls.command_options)

    async def async_settings_menu(self) -> None:
        """
        Options menu on receiver via HTTP get command.

        Only available if using Telnet.
        """
        if self._settings_menu:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_setup_close
            )
        else:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_setup_open
            )

    async def async_channel_level_adjust(self) -> None:
        """Toggle the channel level adjust menu on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_channel_level_adjust
            )
        else:
            await self.api.async_get_command(self.urls.command_channel_level_adjust)

    async def async_dimmer_toggle(self) -> None:
        """Toggle dimmer on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_dimmer_toggle
            )
        else:
            await self.api.async_get_command(self.urls.command_dimmer_toggle)

    async def async_dimmer(self, mode: DimmerModes) -> None:
        """Set dimmer mode on receiver via HTTP get command."""
        if mode not in self._dimmer_modes:
            raise AvrCommandError("Invalid dimmer mode")

        mapped_mode = DIMMER_MODE_MAP[mode]
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_dimmer_set.format(mode=mapped_mode)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_dimmer_set.format(mode=mapped_mode)
            )

    async def async_tactile_transducer_on(self) -> None:
        """Turn on tactile transducer on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_tactile_transducer.format(mode="ON")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_tactile_transducer.format(mode="ON")
            )

    async def async_auto_standby(self, auto_standby: AutoStandbys) -> None:
        """Set auto standby on receiver via HTTP get command."""
        if auto_standby not in self._auto_standbys:
            raise AvrCommandError("Invalid Auto Standby mode")
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_auto_standby.format(mode=auto_standby)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_auto_standby.format(mode=auto_standby)
            )

    async def async_sleep(self, sleep: Union[Literal["OFF"], int]) -> None:
        """
        Set auto standby on receiver via HTTP get command.

        Valid sleep values are "OFF" and 1-120 (in minutes)
        """
        if sleep != "OFF" and sleep not in range(1, 120):
            raise AvrCommandError("Invalid sleep value")

        local_sleep = f"{sleep:03}" if isinstance(sleep, int) else sleep
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_sleep.format(value=local_sleep)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_sleep.format(value=local_sleep)
            )

    async def async_tactile_transducer_off(self) -> None:
        """Turn on tactile transducer on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_tactile_transducer.format(mode="OFF")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_tactile_transducer.format(mode="OFF")
            )

    async def async_tactile_transducer_toggle(self) -> None:
        """
        Turn on tactile transducer on receiver via HTTP get command.

        Only available if using Telnet.
        """
        if self._tactile_transducer != "OFF":
            await self.async_tactile_transducer_off()
        else:
            await self.async_tactile_transducer_on()

    async def async_tactile_transducer_level_up(self) -> None:
        """Increase the transducer level on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_tactile_transducer_level.format(mode="UP")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_tactile_transducer_level.format(mode="UP")
            )

    async def async_tactile_transducer_level_down(self) -> None:
        """Decrease the transducer level on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_tactile_transducer_level.format(
                    mode="DOWN"
                )
            )
        else:
            await self.api.async_get_command(
                self.urls.command_tactile_transducer_level.format(mode="DOWN")
            )

    async def async_transducer_lpf(self, lpf: TransducerLPFs):
        """Set transducer low pass filter on receiver via HTTP get command."""
        if lpf not in self._tactile_transducer_lpfs:
            raise AvrCommandError("Invalid tactile transducer low pass filter")

        frequency = lpf.split()[0]
        if len(frequency) == 2:
            frequency = f"0{frequency}"
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_tactile_transducer_lpf.format(
                    frequency=frequency
                )
            )
        else:
            await self.api.async_get_command(
                self.urls.command_tactile_transducer_lpf.format(frequency=frequency)
            )

    async def async_room_size(self, room_size: RoomSizes) -> None:
        """Set room size on receiver via HTTP get command."""
        if room_size not in self._room_sizes:
            raise AvrCommandError("Invalid room size")

        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_room_size.format(size=room_size)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_room_size.format(size=room_size)
            )

    async def async_trigger_on(self, trigger: int) -> None:
        """
        Set trigger to ON on receiver via HTTP get command.

        :param trigger: Trigger number to set to ON. Valid values are 1-3.
        """
        if trigger < 1 or trigger > 3:
            raise AvrCommandError("Trigger number must be between 1 and 3")

        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_trigger.format(number=trigger, mode="ON")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_trigger.format(number=trigger, mode="ON")
            )

    async def async_trigger_off(self, trigger: int) -> None:
        """
        Set trigger to OFF on receiver via HTTP get command.

        :param trigger: Trigger number to set to OFF. Valid values are 1-3.
        """
        if trigger < 1 or trigger > 3:
            raise AvrCommandError("Trigger number must be between 1 and 3")

        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_trigger.format(number=trigger, mode="OFF")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_trigger.format(number=trigger, mode="OFF")
            )

    async def async_trigger_toggle(self, trigger: int) -> None:
        """
        Toggle trigger on receiver via HTTP get command.

        Only available if using Telnet.

        :param trigger: Trigger number to toggle. Valid values are 1-3.
        """
        if trigger < 1 or trigger > 3:
            raise AvrCommandError("Trigger number must be between 1 and 3")

        trigger_status = self._triggers.get(trigger)
        if trigger_status == "ON":
            await self.async_trigger_off(trigger)
        else:
            await self.async_trigger_on(trigger)

    async def async_quick_select_mode(self, quick_select_number: int) -> None:
        """
        Set quick select mode on receiver via HTTP get command.

        :param quick_select_number: Quick select number to set. Valid values are 1-5.
        """
        if quick_select_number not in range(1, 5):
            raise AvrCommandError("Quick select number must be between 1 and 5")

        if self.telnet_available:
            if "denon" in self.manufacturer.lower():
                command = self.telnet_commands.command_quick_select_memory
            else:
                command = self.telnet_commands.command_smart_select_memory
            await self.telnet_api.async_send_commands(
                command.format(number=quick_select_number)
            )
        else:
            if "denon" in self.manufacturer.lower():
                command = self.urls.command_quick_select_mode
            else:
                command = self.urls.command_smart_select_mode
            await self.api.async_get_command(command.format(number=quick_select_number))

    async def async_quick_select_memory(self, quick_select_number: int) -> None:
        """
        Set quick select memory on receiver via HTTP get command.

        :param quick_select_number: Quick select number to set. Valid values are 1-5.
        """
        if quick_select_number not in range(1, 5):
            raise AvrCommandError("Quick select number must be between 1 and 5")

        if self.telnet_available:
            if "denon" in self.manufacturer.lower():
                command = self.telnet_commands.command_quick_select_mode
            else:
                command = self.telnet_commands.command_smart_select_mode
            await self.telnet_api.async_send_commands(
                command.format(number=quick_select_number)
            )
        else:
            if "denon" in self.manufacturer.lower():
                command = self.urls.command_quick_select_memory
            else:
                command = self.urls.command_smart_select_memory
            await self.api.async_get_command(command.format(number=quick_select_number))

    async def async_delay_up(self) -> None:
        """Delay up on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_delay_up
            )
        else:
            await self.api.async_get_command(self.urls.command_delay_up)

    async def async_delay_down(self) -> None:
        """Delay down on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_delay_down
            )
        else:
            await self.api.async_get_command(self.urls.command_delay_down)

    async def async_eco_mode(self, mode: EcoModes) -> None:
        """Set Eco mode."""
        if mode not in self._eco_modes:
            raise AvrCommandError("Invalid Eco mode")

        mapped_mode = ECO_MODE_MAP[mode]
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_eco_mode.format(mode=mapped_mode)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_eco_mode.format(mode=mapped_mode)
            )

    async def async_hdmi_output(self, output: HDMIOutputs) -> None:
        """Set HDMI output."""
        if output not in self._hdmi_outputs:
            raise AvrCommandError("Invalid HDMI output mode")

        mapped_output = HDMI_OUTPUT_MAP[output]
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_hdmi_output.format(output=mapped_output)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_hdmi_output.format(output=mapped_output)
            )

    async def async_hdmi_audio_decode(self, mode: HDMIAudioDecodes) -> None:
        """Set HDMI Audio Decode mode on receiver via HTTP get command."""
        if mode not in self._hdmi_audio_decodes:
            raise AvrCommandError("Invalid HDMI Audio Decode mode")

        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_hdmi_audio_decode.format(mode=mode)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_hdmi_audio_decode.format(mode=mode)
            )

    async def async_video_processing_mode(self, mode: VideoProcessingModes) -> None:
        """Set video processing mode on receiver via HTTP get command."""
        if mode not in self._video_processing_modes:
            raise AvrCommandError("Invalid video processing mode")
        processing_mode = VIDEO_PROCESSING_MODES_MAP[mode]
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_video_processing_mode.format(
                    mode=processing_mode
                )
            )
        else:
            await self.api.async_get_command(
                self.urls.command_video_processing_mode.format(mode=processing_mode)
            )

    async def async_status(self) -> str:
        """Get status of receiver via HTTP get command."""
        if "denon" not in self.manufacturer.lower():
            raise AvrCommandError("Status command is only supported for Denon devices")
        return await self.api.async_get_command(self.urls.command_status)

    async def async_system_reset(self) -> None:
        """DANGER! Reset the receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_system_reset
            )
        else:
            await self.api.async_get_command(self.urls.command_system_reset)

    async def async_network_restart(self) -> None:
        """Restart the network on the receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_network_restart
            )
        else:
            await self.api.async_get_command(self.urls.command_network_restart)

    async def async_speaker_preset(self, preset: int) -> None:
        """
        Set speaker preset on receiver via HTTP get command.

        Valid preset values are 1-2.
        """
        if preset < 1 or preset > 2:
            raise AvrCommandError("Speaker preset number must be 1 or 2")

        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_speaker_preset.format(number=preset)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_speaker_preset.format(number=preset)
            )

    async def async_speaker_preset_toggle(self) -> None:
        """
        Toggle speaker preset on receiver via HTTP get command.

        Only available if using Telnet.
        """
        speaker_preset = 1 if self._speaker_preset == 2 else 2
        await self.async_speaker_preset(speaker_preset)

    async def async_bt_transmitter_on(
        self,
    ) -> None:
        """Turn on Bluetooth transmitter on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_bluetooth_transmitter.format(mode="ON")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_bluetooth_transmitter.format(mode="ON")
            )

    async def async_bt_transmitter_off(
        self,
    ) -> None:
        """Turn off Bluetooth transmitter on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_bluetooth_transmitter.format(mode="OFF")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_bluetooth_transmitter.format(mode="OFF")
            )

    async def async_bt_transmitter_toggle(self) -> None:
        """
        Toggle Bluetooth transmitter mode on receiver via HTTP get command.

        Only available if using Telnet.
        """
        if self.bt_transmitter:
            await self.async_bt_transmitter_off()
        else:
            await self.async_bt_transmitter_on()

    async def async_bt_output_mode(self, mode: BluetoothOutputModes) -> None:
        """Set Bluetooth transmitter mode on receiver via HTTP get command."""
        if mode not in self._bt_output_modes:
            raise AvrCommandError("Invalid Bluetooth output mode")

        mapped_mode = BLUETOOTH_OUTPUT_MODES_MAP[mode]
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_bluetooth_transmitter.format(
                    mode=mapped_mode
                )
            )
        else:
            await self.api.async_get_command(
                self.urls.command_bluetooth_transmitter.format(mode=mapped_mode)
            )

    async def async_bt_output_mode_toggle(self) -> None:
        """
        Toggle Bluetooth output mode on receiver via HTTP get command.

        Only available if using Telnet.
        """
        if self.bt_output_mode == "Bluetooth + Speakers":
            await self.async_bt_output_mode("Bluetooth Only")
        else:
            await self.async_bt_output_mode("Bluetooth + Speakers")

    async def async_delay_time_up(self) -> None:
        """Delay time up on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_delay_time.format(value="UP")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_delay_time.format(value="UP")
            )

    async def async_delay_time_down(self) -> None:
        """Delay time up on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_delay_time.format(value="DOWN")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_delay_time.format(value="DOWN")
            )

    async def async_delay_time(self, delay_time: int) -> None:
        """
        Set delay time on receiver via HTTP get command.

        :param delay_time: Delay time in ms. Valid values are 0-999.
        """
        if delay_time < 0 or delay_time > 999:
            raise AvrCommandError("Invalid delay time value")

        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_delay_time.format(value=delay_time)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_delay_time.format(value=delay_time)
            )

    async def async_audio_restorer(self, mode: AudioRestorers):
        """Set audio restorer on receiver via HTTP get command."""
        if mode not in self._audio_restorers:
            raise AvrCommandError("Invalid audio restorer mode")

        mapped_mode = AUDIO_RESTORER_MAP[mode]
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_audio_restorer.format(mode=mapped_mode)
            )
        else:
            await self.api.async_get_command(
                self.urls.command_audio_restorer.format(mode=mapped_mode)
            )

    async def async_remote_control_lock(self):
        """Set remote control lock on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_remote_control_lock.format(mode="ON")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_remote_control_lock.format(mode="ON")
            )

    async def async_remote_control_unlock(self):
        """Set remote control unlock on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_remote_control_lock.format(mode="OFF")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_remote_control_lock.format(mode="OFF")
            )

    async def async_panel_lock(self, panel_lock_mode: PanelLocks):
        """Set panel lock on receiver via HTTP get command."""
        if panel_lock_mode not in self._panel_locks:
            raise AvrCommandError("Invalid panel lock mode")

        if self.telnet_available:
            if panel_lock_mode == "Panel":
                await self.telnet_api.async_send_commands(
                    self.telnet_commands.command_panel_lock.format(mode="ON")
                )
            else:
                await self.telnet_api.async_send_commands(
                    self.telnet_commands.command_panel_and_volume_lock
                )
        else:
            if panel_lock_mode == "Panel":
                await self.api.async_get_command(
                    self.urls.command_panel_lock.format(mode="ON")
                )
            else:
                await self.api.async_get_command(
                    self.urls.command_panel_and_volume_lock
                )

    async def async_panel_unlock(self):
        """Set panel unlock on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_panel_lock.format(mode="OFF")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_panel_lock.format(mode="OFF")
            )

    async def async_graphic_eq_on(self) -> None:
        """Turn on Graphic EQ on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_graphic_eq.format(mode="ON")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_graphic_eq.format(mode="ON")
            )

    async def async_graphic_eq_off(self) -> None:
        """Turn off Graphic EQ on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_graphic_eq.format(mode="OFF")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_graphic_eq.format(mode="OFF")
            )

    async def async_graphic_eq_toggle(self) -> None:
        """
        Toggle Graphic EQ on receiver via HTTP get command.

        Only available if using Telnet.
        """
        if self._graphic_eq:
            await self.async_graphic_eq_off()
        else:
            await self.async_graphic_eq_on()

    async def async_headphone_eq_on(self) -> None:
        """Turn on Headphone EQ on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_headphone_eq.format(mode="ON")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_headphone_eq.format(mode="ON")
            )

    async def async_headphone_eq_off(self) -> None:
        """Turn off Headphone EQ on receiver via HTTP get command."""
        if self.telnet_available:
            await self.telnet_api.async_send_commands(
                self.telnet_commands.command_headphone_eq.format(mode="OFF")
            )
        else:
            await self.api.async_get_command(
                self.urls.command_headphone_eq.format(mode="OFF")
            )

    async def async_headphone_eq_toggle(self) -> None:
        """
        Toggle Headphone EQ on receiver via HTTP get command.

        Only available if using Telnet.
        """
        if self._headphone_eq:
            await self.async_headphone_eq_off()
        else:
            await self.async_headphone_eq_on()


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRFoundation:
    """
    Implements the foundation class of DenonAVR functions.

    All functions like the receiver, zones, sound modes, media settings, tone
    control use this class.
    """

    _device: DenonAVRDeviceInfo = attr.ib(
        validator=attr.validators.instance_of(DenonAVRDeviceInfo),
        default=attr.Factory(DenonAVRDeviceInfo),
        kw_only=True,
    )
    _is_setup: bool = attr.ib(converter=bool, default=False, init=False)

    async def async_update_attrs_appcommand(
        self,
        update_attrs: Dict[AppCommandCmd, None],
        appcommand0300: bool = False,
        global_update: bool = False,
        cache_id: Optional[Hashable] = None,
    ):
        """Update attributes from AppCommand.xml."""
        # Copy that we do not accidently change the wrong dict
        update_attrs = deepcopy(update_attrs)
        # Collect tags for AppCommand.xml call
        tags = tuple(i for i in update_attrs.keys())
        # Execute call
        try:
            if global_update:
                xml = await self._device.api.async_get_global_appcommand(
                    appcommand0300=appcommand0300, cache_id=cache_id
                )
            else:
                # Determine endpoint
                if appcommand0300:
                    url = self._device.urls.appcommand0300
                else:
                    url = self._device.urls.appcommand
                xml = await self._device.api.async_post_appcommand(
                    url, tags, cache_id=cache_id
                )
        except AvrRequestError as err:
            _LOGGER.debug("Error when getting status update: %s", err)
            raise

        # Extract relevant information
        zone = self._device.get_own_zone()

        attrs = deepcopy(update_attrs)
        for app_command in attrs.keys():
            search_strings = self.create_appcommand_search_strings(app_command, zone)
            start = 0
            success = 0
            for i, pattern in enumerate(app_command.response_pattern):
                try:
                    start += 1
                    # Check if attribute exists
                    getattr(self, pattern.update_attribute)
                    # Set new value either from XML attribute or text
                    if pattern.get_xml_attribute is not None:
                        set_value = xml.find(search_strings[i]).get(
                            pattern.get_xml_attribute
                        )
                    else:
                        set_value = xml.find(search_strings[i]).text

                    setattr(self, pattern.update_attribute, set_value)
                    success += 1

                    _LOGGER.debug(
                        "Changing variable %s to value %s",
                        pattern.update_attribute,
                        set_value,
                    )

                except (AttributeError, IndexError) as err:
                    _LOGGER.debug(
                        "Failed updating attribute %s for zone %s: %s",
                        pattern.update_attribute,
                        self._device.zone,
                        err,
                    )

            if start == success:
                # Done
                update_attrs.pop(app_command, None)

        # Check if each attribute was updated
        if update_attrs:
            raise AvrProcessingError(
                f"Some attributes of zone {self._device.zone} not found on update:"
                f" {update_attrs}"
            )

    async def async_update_attrs_status_xml(
        self,
        update_attrs: Dict[str, str],
        urls: List[str],
        cache_id: Optional[Hashable] = None,
    ):
        """
        Update attributes from status xml.

        # URLs to be scanned. Like:
        urls = [self._device.urls.status, self._device.urls.mainzone]

        # Variables with their tags to be updated.
        # Key = Variable, Value = XML tag Like:
        update_attrs = {"power": "./Power/value"}
        """
        # Copy that we do not accidently change the wrong dict
        update_attrs = deepcopy(update_attrs)

        for url in urls:
            try:
                xml = await self._device.api.async_get_xml(url, cache_id=cache_id)
            except AvrRequestError as err:
                _LOGGER.debug(
                    "Error when getting status update from url %s: %s", url, err
                )
                continue
            attrs = deepcopy(update_attrs)
            for name, tag in attrs.items():
                try:
                    # Check if attribute exists
                    getattr(self, name)
                    # Set new value
                    setattr(self, name, xml.find(tag).text)
                    # Done
                    update_attrs.pop(name, None)

                    _LOGGER.debug(
                        "Changing variable %s to value %s", name, xml.find(tag).text
                    )

                except (AttributeError, IndexError) as err:
                    _LOGGER.debug(
                        "Failed updating attribute %s for zone %s: %s",
                        name,
                        self._device.zone,
                        err,
                    )

            # All done, no need for continuing
            if not update_attrs:
                break

        # Check if each attribute was updated
        if update_attrs:
            raise AvrProcessingError(
                f"Some attributes of zone {self._device.zone} not found on update:"
                f" {update_attrs}"
            )

    @staticmethod
    def create_appcommand_search_strings(
        app_command_cmd: AppCommandCmd, zone: str
    ) -> List[str]:
        """Create search pattern for AppCommand(0300).xml response."""
        result = []

        for resp in app_command_cmd.response_pattern:
            string = "./cmd"
            # Text of cmd tag in query was added as attribute to response
            if app_command_cmd.cmd_text:
                string = (
                    string + f"[@{APPCOMMAND_CMD_TEXT}='{app_command_cmd.cmd_text}']"
                )
            # Text of name tag in query was added as attribute to response
            if app_command_cmd.name:
                string = string + f"[@{APPCOMMAND_NAME}='{app_command_cmd.name}']"
            # Some results include a zone tag
            if resp.add_zone:
                string = string + f"/{zone}"
            # Suffix like /status, /volume
            string = string + resp.suffix

            # A complete search string with all strributes set looks like
            # ./cmd[@cmd_text={cmd_text}][@name={name}]/zone1/volume
            result.append(string)

        return result


def set_api_host(
    instance: DenonAVRFoundation, attribute: attr.Attribute, value: str
) -> str:
    """Change API host on host changes too."""
    # First change _device.api.host then return value
    # pylint: disable=protected-access
    instance._device.api.host = value
    instance._device.telnet_api.host = value
    return value


def set_api_timeout(
    instance: DenonAVRFoundation, attribute: attr.Attribute, value: float
) -> float:
    """Change API timeout on timeout changes too."""
    # First change _device.api.host then return value
    # pylint: disable=protected-access
    instance._device.api.timeout = value
    instance._device.api.read_timeout = max(value, 15.0)
    instance._device.telnet_api.timeout = value
    instance._device.telnet_api._send_confirmation_timeout = value
    return value


def convert_string_int_bool(value: Union[str, bool]) -> bool:
    """Convert an integer from string format to bool."""
    if value is None:
        return None
    if isinstance(value, bool):
        return value
    return bool(int(value))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module implements the handler for sound mode of Denon AVR receivers.

:copyright: (c) 2021 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import asyncio
import logging
from collections.abc import Hashable
from copy import deepcopy
from typing import Dict, List, Literal, Optional, get_args

import attr

from .appcommand import AppCommands
from .const import (
    ALL_ZONE_STEREO,
    AURO_3D_MODE_MAP,
    AURO_3D_MODE_MAP_MAP_LABELS,
    AURO_MATIC_3D_PRESET_MAP,
    AURO_MATIC_3D_PRESET_MAP_LABELS,
    DENON_ATTR_SETATTR,
    DIALOG_ENHANCER_LEVEL_MAP,
    DIALOG_ENHANCER_LEVEL_MAP_LABELS,
    EFFECT_SPEAKER_SELECTION_MAP,
    EFFECT_SPEAKER_SELECTION_MAP_LABELS,
    SOUND_MODE_MAPPING,
    Auro3DModes,
    AuroMatic3DPresets,
    DialogEnhancerLevels,
    DRCs,
    EffectSpeakers,
    IMAXHPFs,
    IMAXLPFs,
)
from .exceptions import AvrCommandError, AvrIncompleteResponseError, AvrProcessingError
from .foundation import DenonAVRFoundation, convert_on_off_bool

_LOGGER = logging.getLogger(__name__)


def convert_sound_mode(value: Optional[str]) -> Optional[str]:
    """Remove multiple spaces from value."""
    if value is None:
        return value
    return " ".join(str(value).split())


def sound_mode_map_factory() -> Dict[str, List]:
    """Construct sound_mode map."""
    sound_mode_map = {}
    for matched_mode, sublist in SOUND_MODE_MAPPING.items():
        sound_mode_map[matched_mode] = [convert_sound_mode(i) for i in sublist]
    return sound_mode_map


def sound_mode_rev_map_factory(instance: "DenonAVRSoundMode") -> Dict[str, str]:
    """
    Construct the sound_mode_rev_map.

    Reverse the key value structure. The sound_mode_rev_map is bigger,
    but allows for direct matching using a dictionary key access.
    The sound_mode_map is uses externally to set this dictionary
    because that has a nicer syntax.
    """
    mode_map = list(
        instance._sound_mode_map.items()  # pylint: disable=protected-access
    )
    mode_map_rev = {}
    for matched_mode, sublist in mode_map:
        for raw_mode in sublist:
            mode_map_rev[convert_sound_mode(raw_mode)] = matched_mode
    return mode_map_rev


@attr.s(auto_attribs=True, on_setattr=DENON_ATTR_SETATTR)
class DenonAVRSoundMode(DenonAVRFoundation):
    """This class implements sound mode functions of Denon AVR receiver."""

    _support_sound_mode: Optional[bool] = attr.ib(
        converter=attr.converters.optional(bool), default=None
    )
    _sound_mode_raw: Optional[str] = attr.ib(
        converter=attr.converters.optional(convert_sound_mode), default=None
    )
    _neural_x: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _imax_auto_off: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _imax_audio_settings: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _imax_hpf: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    _imax_hpfs = get_args(IMAXHPFs)
    _imax_lpf: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    _imax_lpfs = get_args(IMAXLPFs)
    _imax_subwoofer_mode: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _imax_subwoofer_output: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _cinema_eq: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _center_spread: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _loudness_management: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _dialog_enhancer_level: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _dialog_enhancer_levels = get_args(DialogEnhancerLevels)
    _auromatic_3d_preset: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _auromatic_3d_presets = get_args(AuroMatic3DPresets)
    _auromatic_3d_strength: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    _auro_3d_mode: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _auro_3d_modes = get_args(Auro3DModes)
    _dialog_control: Optional[int] = attr.ib(
        converter=attr.converters.optional(int), default=None
    )
    _speaker_virtualizer: Optional[bool] = attr.ib(
        converter=attr.converters.optional(convert_on_off_bool), default=None
    )
    _effect_speaker_selection: Optional[str] = attr.ib(
        converter=attr.converters.optional(str), default=None
    )
    _effect_speakers = get_args(EffectSpeakers)
    _drc: Optional[str] = attr.ib(converter=attr.converters.optional(str), default=None)
    _drcs = get_args(DRCs)
    _sound_mode_map: Dict[str, list] = attr.ib(
        validator=attr.validators.deep_mapping(
            attr.validators.instance_of(str),
            attr.validators.deep_iterable(
                attr.validators.instance_of(str), attr.validators.instance_of(list)
            ),
            attr.validators.instance_of(dict),
        ),
        default=attr.Factory(sound_mode_map_factory),
        init=False,
    )
    _sound_mode_map_rev: Dict[str, str] = attr.ib(
        validator=attr.validators.deep_mapping(
            attr.validators.instance_of(str),
            attr.validators.instance_of(str),
            attr.validators.instance_of(dict),
        ),
        default=attr.Factory(sound_mode_rev_map_factory, takes_self=True),
        init=False,
    )
    _setup_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))
    _appcommand_active: bool = attr.ib(converter=bool, default=True, init=False)

    # Update tags for attributes
    # AppCommand.xml interface
    appcommand_attrs = {AppCommands.GetSurroundModeStatus: None}
    # Status.xml interface
    status_xml_attrs_01 = {"_sound_mode_raw": "./selectSurround/value"}
    status_xml_attrs_02 = {"_sound_mode_raw": "./SurrMode/value"}

    async def async_setup(self) -> None:
        """Ensure that the instance is initialized."""
        async with self._setup_lock:
            _LOGGER.debug("Starting sound mode setup")

            # The first update determines if sound mode is supported
            await self.async_update_sound_mode()

            if self._support_sound_mode and self._appcommand_active:
                # Add tags for a potential AppCommand.xml update
                for tag in self.appcommand_attrs:
                    self._device.api.add_appcommand_update_tag(tag)

            self._device.telnet_api.register_callback(
                "MS", self._async_soundmode_callback
            )
            self._device.telnet_api.register_callback(
                "PS", self._async_neural_x_callback
            )
            self._device.telnet_api.register_callback("PS", self._async_imax_callback)
            self._device.telnet_api.register_callback(
                "PS", self._async_cinema_eq_callback
            )
            self._device.telnet_api.register_callback(
                "PS", self._async_center_spread_callback
            )
            self._device.telnet_api.register_callback(
                "PS", self._async_loudness_management_callback
            )
            self._device.telnet_api.register_callback(
                "PS", self._async_dialog_enhancer_callback
            )
            self._device.telnet_api.register_callback("PS", self._async_auro_callback)
            self._device.telnet_api.register_callback(
                "PS", self._async_dialog_control_callback
            )
            self._device.telnet_api.register_callback(
                "PS", self._async_speaker_virtualizer_callback
            )
            self._device.telnet_api.register_callback(
                "PS", self._async_effect_speaker_selection_callback
            )
            self._device.telnet_api.register_callback("PS", self._async_drc_callback)

            self._is_setup = True
            _LOGGER.debug("Finished sound mode setup")

    async def _async_soundmode_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a sound mode change event."""
        if self._device.zone != zone:
            return

        # QUICK/SMART are not sound modes
        if parameter.startswith("QUICK") or parameter.startswith("SMART"):
            return

        self._sound_mode_raw = parameter

    async def _async_neural_x_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Neural X:change event."""
        parameter_name_length = len("NEURAL")
        if parameter[:parameter_name_length] == "NEURAL":
            self._neural_x = parameter[parameter_name_length + 1 :]

    async def _async_imax_callback(self, zone: str, event: str, parameter: str) -> None:
        """Handle an IMAX change event."""
        key_value = parameter.split()
        if len(key_value) != 2 or key_value[0][:4] != "IMAX":
            return

        if key_value[0] == "IMAX":
            self._imax_auto_off = parameter[5:]
        elif key_value[0] == "IMAXAUD":
            self._imax_audio_settings = parameter[8:]
        elif key_value[0] == "IMAXHPF":
            self._imax_hpf = int(parameter[8:])
        elif key_value[0] == "IMAXLPF":
            self._imax_lpf = int(parameter[8:])
        elif key_value[0] == "IMAXSWM":
            self._imax_subwoofer_mode = parameter[8:]
        elif key_value[0] == "IMAXSWO":
            self._imax_subwoofer_output = parameter[8:]

    async def _async_cinema_eq_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Cinema EQ change event."""
        if parameter[:10] == "CINEMA EQ.":
            self._cinema_eq = parameter[10:]

    async def _async_center_spread_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Center Spread change event."""
        if parameter[:3] == "CES":
            self._center_spread = parameter[4:]

    async def _async_loudness_management_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Loudness Management change event."""
        if parameter[:3] == "LOM":
            self._loudness_management = parameter[4:]

    async def _async_dialog_enhancer_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Dialog Enhancer change event."""
        if parameter[:3] == "DEH":
            self._dialog_enhancer_level = DIALOG_ENHANCER_LEVEL_MAP_LABELS[
                parameter[4:]
            ]

    async def _async_auro_callback(self, zone: str, event: str, parameter: str) -> None:
        """Handle a Auro change event."""
        key_value = parameter.split()
        if len(key_value) != 2 or key_value[0][:4] != "AURO":
            return

        if key_value[0] == "AUROPR":
            self._auromatic_3d_preset = AURO_MATIC_3D_PRESET_MAP_LABELS[parameter[7:]]
        elif key_value[0] == "AUROST":
            self._auromatic_3d_strength = int(parameter[7:])
        elif key_value[0] == "AUROMODE":
            self._auro_3d_mode = AURO_3D_MODE_MAP_MAP_LABELS[parameter[9:]]

    async def _async_dialog_control_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Dialog Control change event."""
        key_value = parameter.split()
        if len(key_value) != 2 or key_value[0] != "DIC":
            return

        self._dialog_control = int(key_value[1])

    async def _async_speaker_virtualizer_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Speaker Virtualizer change event."""
        key_value = parameter.split()
        if len(key_value) != 2 or key_value[0] != "SPV":
            return

        self._speaker_virtualizer = key_value[1]

    async def _async_effect_speaker_selection_callback(
        self, zone: str, event: str, parameter: str
    ) -> None:
        """Handle a Effect Speaker Selection change event."""
        key_value = parameter.split(":")
        if len(key_value) != 2 or key_value[0] != "SP":
            return

        self._effect_speaker_selection = EFFECT_SPEAKER_SELECTION_MAP_LABELS[
            key_value[1]
        ]

    async def _async_drc_callback(self, zone: str, event: str, parameter: str) -> None:
        """Handle a DRC change event."""
        key_value = parameter.split()
        if len(key_value) != 2 or key_value[0] != "DRC":
            return

        self._drc = key_value[1]

    async def async_update(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ) -> None:
        """Update sound mode asynchronously."""
        _LOGGER.debug("Starting sound mode update")
        # Ensure instance is setup before updating
        if not self._is_setup:
            await self.async_setup()

        # Update state
        await self.async_update_sound_mode(
            global_update=global_update, cache_id=cache_id
        )
        _LOGGER.debug("Finished sound mode update")

    async def async_update_sound_mode(
        self, global_update: bool = False, cache_id: Optional[Hashable] = None
    ):
        """Update sound mode status of device."""
        if self._device.use_avr_2016_update is None:
            raise AvrProcessingError(
                "Device is not setup correctly, update method not set"
            )

        if self._is_setup and not self._support_sound_mode:
            return

        if self._device.use_avr_2016_update and self._appcommand_active:
            try:
                await self.async_update_attrs_appcommand(
                    self.appcommand_attrs,
                    global_update=global_update,
                    cache_id=cache_id,
                )
            except (AvrProcessingError, AvrIncompleteResponseError):
                self._appcommand_active = False
                _LOGGER.debug(
                    "Appcommand.xml does not support Sound mode. "
                    "Testing status.xml interface next"
                )
            else:
                if not self._is_setup:
                    self._support_sound_mode = True
                    _LOGGER.info("Sound mode supported")
                return

        urls = [self._device.urls.status, self._device.urls.mainzone]
        # There are two different options of sound mode tags
        try:
            await self.async_update_attrs_status_xml(
                self.status_xml_attrs_01, urls, cache_id=cache_id
            )
        except AvrProcessingError:
            try:
                await self.async_update_attrs_status_xml(
                    self.status_xml_attrs_02, urls, cache_id=cache_id
                )
            except AvrProcessingError:
                self._support_sound_mode = False
                _LOGGER.info("Sound mode not supported")
                return

        if not self._is_setup:
            self._support_sound_mode = True
            _LOGGER.info("Sound mode supported")

    def match_sound_mode(self) -> Optional[str]:
        """Match the raw_sound_mode to its corresponding sound_mode."""
        if self._sound_mode_raw is None:
            return None
        smr_up = self._sound_mode_raw.upper()
        try:
            sound_mode = self._sound_mode_map_rev[smr_up]
        except KeyError:
            # Estimate sound mode for unclassified input
            if smr_up.find("DTS") != -1:
                self._sound_mode_map["DTS SURROUND"].append(smr_up)
                _LOGGER.warning(
                    "Not able to match sound mode: '%s', assuming 'DTS SURROUND'.",
                    smr_up,
                )
            elif smr_up.find("DOLBY") != -1:
                self._sound_mode_map["DOLBY DIGITAL"].append(smr_up)
                _LOGGER.warning(
                    "Not able to match sound mode: '%s', assuming 'DOLBY DIGITAL'.",
                    smr_up,
                )
            elif smr_up.find("MUSIC") != -1:
                self._sound_mode_map["MUSIC"].append(smr_up)
                _LOGGER.warning(
                    "Not able to match sound mode: '%s', assuming 'MUSIC'.", smr_up
                )
            elif smr_up.find("AURO") != -1:
                self._sound_mode_map["AURO3D"].append(smr_up)
                _LOGGER.warning(
                    "Not able to match sound mode: '%s', assuming 'AURO3D'.", smr_up
                )
            elif smr_up.find("MOVIE") != -1 or smr_up.find("CINEMA") != -1:
                self._sound_mode_map["MOVIE"].append(smr_up)
                _LOGGER.warning(
                    "Not able to match sound mode: '%s', assuming 'MOVIE'.", smr_up
                )
            else:
                self._sound_mode_map[smr_up] = [smr_up]
                _LOGGER.warning(
                    "Not able to match sound mode: '%s', returning raw sound mode.",
                    smr_up,
                )
            self._sound_mode_map_rev = sound_mode_rev_map_factory(self)
            sound_mode = self._sound_mode_map_rev[smr_up]

        return sound_mode

    async def _async_set_all_zone_stereo(self, zst_on: bool) -> None:
        """
        Set All Zone Stereo option on the device.

        Calls command to activate/deactivate the mode
        """
        command_url = self._device.urls.command_set_all_zone_stereo
        telnet_command = self._device.telnet_commands.command_set_all_zone_stereo
        if zst_on:
            command_url += "ZST ON"
            telnet_command += "ZST ON"
        else:
            command_url += "ZST OFF"
            telnet_command += "ZST OFF"
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self._device.api.async_get_command(command_url)

    ##############
    # Properties #
    ##############
    @property
    def support_sound_mode(self) -> Optional[bool]:
        """Return True if sound mode is supported."""
        return self._support_sound_mode

    @property
    def sound_mode(self) -> Optional[str]:
        """Return the matched current sound mode as a string."""
        sound_mode_matched = self.match_sound_mode()
        return sound_mode_matched

    @property
    def sound_mode_list(self) -> List[str]:
        """Return a list of available sound modes as string."""
        return list(self._sound_mode_map.keys())

    @property
    def sound_mode_map(self) -> Dict[str, str]:
        """Return a dict of available sound modes with their mapping values."""
        return deepcopy(self._sound_mode_map)

    @property
    def sound_mode_map_rev(self) -> Dict[str, str]:
        """Return a dict to map each sound_mode_raw to matching sound_mode."""
        return deepcopy(self._sound_mode_map_rev)

    @property
    def sound_mode_raw(self) -> Optional[str]:
        """Return the current sound mode as string as received from the AVR."""
        return self._sound_mode_raw

    @property
    def neural_x(self) -> Optional[bool]:
        """
        Return the current Neural:X status.

        Only available if using Telnet.
        """
        return self._neural_x

    @property
    def imax(self) -> Optional[str]:
        """
        Return the current IMAX status.

        Only available if using Telnet.

        Possible values are: "AUTO", "OFF"
        """
        return self._imax_auto_off

    @property
    def imax_audio_settings(self) -> Optional[str]:
        """
        Return the current IMAX Audio Settings.

        Only available if using Telnet.

        Possible values are: "AUTO", "MANUAL"
        """
        return self._imax_audio_settings

    @property
    def imax_hpf(self) -> Optional[int]:
        """
        Return the current IMAX High Pass Filter.

        Only available if using Telnet.
        """
        return self._imax_hpf

    @property
    def imax_lpf(self) -> Optional[int]:
        """
        Return the current IMAX Low Pass Filter.

        Only available if using Telnet.
        """
        return self._imax_lpf

    @property
    def imax_subwoofer_mode(self) -> Optional[str]:
        """
        Return the current IMAX Subwoofer Mode.

        Only available if using Telnet.
        """
        return self._imax_subwoofer_mode

    @property
    def imax_subwoofer_output(self) -> Optional[str]:
        """
        Return the current IMAX Subwoofer Output Mode.

        Only available if using Telnet.
        """
        return self._imax_subwoofer_output

    @property
    def cinema_eq(self) -> Optional[bool]:
        """
        Return the current Cinema EQ status.

        Only available if using Telnet.
        """
        return self._cinema_eq

    @property
    def center_spread(self) -> Optional[bool]:
        """
        Return the current Center Spread status.

        Only available if using Telnet.
        """
        return self._center_spread

    @property
    def loudness_management(self) -> Optional[bool]:
        """
        Return the current Loudness Management status.

        Only available if using Telnet.
        """
        return self._loudness_management

    @property
    def dialog_enhancer(self) -> Optional[str]:
        """
        Return the current Dialog Enhancer level.

        Only available if using Telnet.

        Possible values are: "Off", "Low", "Medium", "High"
        """
        return self._dialog_enhancer_level

    @property
    def auromatic_3d_preset(self) -> Optional[str]:
        """
        Return the current Auro-Matic 3D Preset.

        Only available if using Telnet.

        Possible values are: "Small, "Medium", "Large", "Speech", "Movie"
        """
        return self._auromatic_3d_preset

    @property
    def auromatic_3d_strength(self) -> Optional[int]:
        """
        Return the current Auro-Matic 3D Strength.

        Only available if using Telnet.

        Possible values are: 1-16
        """
        return self._auromatic_3d_strength

    @property
    def auro_3d_mode(self) -> Optional[str]:
        """
        Return the current Auro 3D mode.

        Only available if using Telnet.

        Possible values are: "Direct", "Channel Expansion"
        """
        return self._auro_3d_mode

    @property
    def dialog_control(self) -> Optional[int]:
        """
        Return the current Dialog Control level.

        Only available if using Telnet.

        Possible values are: 0-6
        """
        return self._dialog_control

    @property
    def speaker_virtualizer(self) -> Optional[bool]:
        """
        Return the current Speaker Virtualizer status.

        Only available if using Telnet.
        """
        return self._speaker_virtualizer

    @property
    def effect_speaker_selection(self) -> Optional[str]:
        """
        Return the current Effect Speaker Selection.

        Only available if using Telnet.

        Possible values are: "Floor", "Height + Floor"
        """
        return self._effect_speaker_selection

    @property
    def drc(self) -> Optional[str]:
        """
        Return the current DRC status.

        Only available if using Telnet.

        Possible values are: "AUTO", "LOW", "MID", "HI", "OFF"
        """
        return self._drc

    ##########
    # Setter #
    ##########
    async def async_set_sound_mode(self, sound_mode: str) -> None:
        """
        Set sound_mode of device.

        Valid values depend on the device and should be taken from
        "sound_mode_list".
        """
        if sound_mode not in self.sound_mode_list:
            raise AvrCommandError(f"{sound_mode} is not a valid sound mode")

        if sound_mode == ALL_ZONE_STEREO:
            await self._async_set_all_zone_stereo(True)
            return

        if self.sound_mode == ALL_ZONE_STEREO:
            await self._async_set_all_zone_stereo(False)
        # For selection of sound mode other names then at receiving sound modes
        # have to be used
        # Therefore source mapping is needed to get sound_mode
        # Create command URL and send command via HTTP GET
        command_url = self._device.urls.command_sel_sound_mode + sound_mode
        telnet_command = (
            self._device.telnet_commands.command_sel_sound_mode + sound_mode
        )
        # sent command
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(telnet_command)
        else:
            await self._device.api.async_get_command(command_url)

    async def async_sound_mode_next(self):
        """Select next sound mode."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_sel_sound_mode + "RIGHT"
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_sel_sound_mode + "RIGHT"
            )

    async def async_sound_mode_previous(self):
        """Select previous sound mode."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_sel_sound_mode + "LEFT"
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_sel_sound_mode + "LEFT"
            )

    async def async_neural_x_on(self):
        """Turn on Neural:X sound mode."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_neural_x_on_off.format(mode="ON")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_neural_x_on_off.format(mode="ON")
            )

    async def async_neural_x_off(self):
        """Turn off Neural:X sound mode."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_neural_x_on_off.format(mode="OFF")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_neural_x_on_off.format(mode="OFF")
            )

    async def async_neural_x_toggle(self):
        """
        Toggle Neural:X sound mode.

        Only available if using Telnet.
        """
        if self._neural_x:
            await self.async_neural_x_off()
        else:
            await self.async_neural_x_on()

    async def async_imax_auto(self):
        """Set IMAX sound mode to Auto."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_imax_auto_off.format(mode="AUTO")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_imax_auto_off.format(mode="AUTO")
            )

    async def async_imax_off(self):
        """Turn off IMAX sound mode."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_imax_auto_off.format(mode="OFF")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_imax_auto_off.format(mode="OFF")
            )

    async def async_imax_toggle(self):
        """
        Toggle IMAX sound mode between auto and off.

        Only available if using Telnet.
        """
        if self._imax_auto_off != "OFF":
            await self.async_imax_off()
        else:
            await self.async_imax_auto()

    async def async_imax_audio_settings(self, mode: Literal["AUTO", "MANUAL"]):
        """Set IMAX audio settings."""
        if mode not in ["AUTO", "MANUAL"]:
            raise AvrCommandError(f"{mode} is not a valid IMAX audio setting")

        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_imax_audio_settings.format(
                    mode=mode
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_imax_audio_settings.format(mode=mode)
            )

    async def async_imax_audio_settings_toggle(self) -> None:
        """
        Toggle IMAX audio settings between auto and manual.

        Only available if using Telnet.
        """
        if self._imax_audio_settings == "AUTO":
            await self.async_imax_audio_settings("MANUAL")
        else:
            await self.async_imax_audio_settings("AUTO")

    async def async_imax_hpf(self, hpf: IMAXHPFs) -> None:
        """Set IMAX High Pass Filter."""
        if hpf not in self._imax_hpfs:
            raise AvrCommandError(f"{hpf} is not a valid IMAX high pass filter")

        local_hpf = self._padded_pass_filter(hpf)
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_imax_hpf.format(
                    frequency=local_hpf
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_imax_hpf.format(frequency=local_hpf)
            )

    async def async_imax_lpf(self, lpf: IMAXLPFs) -> None:
        """Set IMAX Low Pass Filter."""
        if lpf not in self._imax_lpfs:
            raise AvrCommandError(f"{lpf} is not a valid IMAX low pass filter")

        local_lpf = self._padded_pass_filter(lpf)
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_imax_lpf.format(
                    frequency=local_lpf
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_imax_lpf.format(frequency=local_lpf)
            )

    @staticmethod
    def _padded_pass_filter(pass_filter: str) -> str:
        return f"0{pass_filter}" if len(str(pass_filter)) == 2 else str(pass_filter)

    async def async_imax_subwoofer_mode(self, mode: Literal["ON", "OFF"]) -> None:
        """Set IMAX Subwoofer Mode."""
        if mode not in ["ON", "OFF"]:
            raise AvrCommandError(f"{mode} is not a valid IMAX subwoofer mode")

        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_imax_subwoofer_mode.format(
                    mode=mode
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_imax_subwoofer_mode.format(mode=mode)
            )

    async def async_imax_subwoofer_output(self, mode: Literal["L+M", "LFE"]) -> None:
        """Set IMAX Subwoofer Output Mode."""
        if mode not in ["L+M", "LFE"]:
            raise AvrCommandError(f"{mode} is not a valid IMAX subwoofer output mode")

        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_imax_subwoofer_output.format(
                    mode=mode
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_imax_subwoofer_output.format(mode=mode)
            )

    async def async_cinema_eq_on(self):
        """Set Cinema EQ to ON."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_cinema_eq.format(mode="ON")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_cinema_eq.format(mode="ON")
            )

    async def async_cinema_eq_off(self):
        """Set Cinema EQ to OFF."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_cinema_eq.format(mode="OFF")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_cinema_eq.format(mode="OFF")
            )

    async def async_cinema_eq_toggle(self):
        """
        Toggle Cinema EQ.

        Only available if using Telnet.
        """
        if self._cinema_eq:
            await self.async_cinema_eq_off()
        else:
            await self.async_cinema_eq_on()

    async def async_center_spread_on(self):
        """Set Center Spread to ON."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_center_spread.format(mode="ON")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_center_spread.format(mode="ON")
            )

    async def async_center_spread_off(self):
        """Set Center Spread to ON."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_center_spread.format(mode="OFF")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_center_spread.format(mode="OFF")
            )

    async def async_center_spread_toggle(self):
        """
        Toggle Center Spread.

        Only available if using Telnet.
        """
        if self._center_spread:
            await self.async_center_spread_off()
        else:
            await self.async_center_spread_on()

    async def async_loudness_management_on(self):
        """Set Loudness Management to ON."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_loudness_management.format(
                    mode="ON"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_loudness_management.format(mode="ON")
            )

    async def async_loudness_management_off(self):
        """Set Loudness Management to OFF."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_loudness_management.format(
                    mode="OFF"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_loudness_management.format(mode="OFF")
            )

    async def async_loudness_management_toggle(self):
        """
        Toggle Loudness Management.

        Only available if using Telnet.
        """
        if self._loudness_management:
            await self.async_loudness_management_off()
        else:
            await self.async_loudness_management_on()

    async def async_dialog_enhancer(self, level: DialogEnhancerLevels) -> None:
        """Set Dialog Enhancer level."""
        if level not in self._dialog_enhancer_levels:
            raise AvrCommandError(f"{level} is not a valid dialog enhancer level")

        level_mapped = DIALOG_ENHANCER_LEVEL_MAP[level]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_dialog_enhancer.format(
                    level=level_mapped
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_dialog_enhancer.format(level=level_mapped)
            )

    async def async_auromatic_3d_preset(self, preset: AuroMatic3DPresets) -> None:
        """Set Auro-Matic 3D Preset."""
        if preset not in self._auromatic_3d_presets:
            raise AvrCommandError(f"{preset} is not a valid Auro-Matic 3D Preset")

        local_preset = AURO_MATIC_3D_PRESET_MAP[preset]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_auromatic_3d_preset.format(
                    preset=local_preset
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_auromatic_3d_preset.format(
                    preset=local_preset
                )
            )

    async def async_auromatic_3d_strength_up(self) -> None:
        """Increase Auro-Matic 3D Strength."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_auromatic_3d_strength.format(
                    value="UP"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_auromatic_3d_strength.format(value="UP")
            )

    async def async_auromatic_3d_strength_down(self) -> None:
        """Decrease Auro-Matic 3D Strength."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_auromatic_3d_strength.format(
                    value="DOWN"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_auromatic_3d_strength.format(value="DOWN")
            )

    async def async_auromatic_3d_strength(self, strength: int) -> None:
        """
        Set Auro-Matic 3D Strength.

        :param strength: Strength value to set. Valid values are 1-16.
        """
        if strength < 1 or strength > 16:
            raise AvrCommandError(f"{strength} is not a valid Auro-Matic 3D Strength")

        local_strength = f"{strength:02}"
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_auromatic_3d_strength.format(
                    value=local_strength
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_auromatic_3d_strength.format(
                    value=local_strength
                )
            )

    async def async_auro_3d_mode(self, mode: Auro3DModes) -> None:
        """Set Auro 3D Mode."""
        if mode not in self._auro_3d_modes:
            raise AvrCommandError(f"{mode} is not a valid Auro 3D Mode")

        local_mode = AURO_3D_MODE_MAP[mode]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_auro_3d_mode.format(
                    mode=local_mode
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_auro_3d_mode.format(mode=local_mode)
            )

    async def async_dialog_control_up(self) -> None:
        """Increase Dialog Control level."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_dialog_control.format(value="UP")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_dialog_control.format(value="UP")
            )

    async def async_dialog_control_down(self) -> None:
        """Decrease Dialog Control level."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_dialog_control.format(value="DOWN")
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_dialog_control.format(value="DOWN")
            )

    async def async_dialog_control(self, level: int) -> None:
        """
        Set Dialog Control level.

        :param level: Level to set. Valid values are 0-6.
        """
        if level < 0 or level > 6:
            raise AvrCommandError(f"{level} is not a valid dialog control level")

        local_level = f"{level:02}"
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_dialog_control.format(
                    value=local_level
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_dialog_control.format(value=local_level)
            )

    async def async_speaker_virtualizer_on(self) -> None:
        """Set Speaker Virtualizer to ON."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_speaker_virtualizer.format(
                    mode="ON"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_speaker_virtualizer.format(mode="ON")
            )

    async def async_speaker_virtualizer_off(self) -> None:
        """Set Speaker Virtualizer to OFF."""
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_speaker_virtualizer.format(
                    mode="OFF"
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_speaker_virtualizer.format(mode="OFF")
            )

    async def async_speaker_virtualizer_toggle(self) -> None:
        """
        Toggle Speaker Virtualizer.

        Only available if using Telnet.
        """
        if self._speaker_virtualizer:
            await self.async_speaker_virtualizer_off()
        else:
            await self.async_speaker_virtualizer_on()

    async def async_effect_speaker_selection(self, mode: EffectSpeakers) -> None:
        """Set Effect Speaker."""
        if mode not in self._effect_speakers:
            raise AvrCommandError(f"{mode} is not a valid effect speaker selection")

        local_mode = EFFECT_SPEAKER_SELECTION_MAP[mode]
        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_effect_speaker_selection.format(
                    mode=local_mode
                )
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_effect_speaker_selection.format(
                    mode=local_mode
                )
            )

    async def async_drc(self, mode: DRCs) -> None:
        """Set DRC mode."""
        if mode not in self._drcs:
            raise AvrCommandError(f"{mode} is not a valid DRC mode")

        if self._device.telnet_available:
            await self._device.telnet_api.async_send_commands(
                self._device.telnet_commands.command_drc.format(mode=mode)
            )
        else:
            await self._device.api.async_get_command(
                self._device.urls.command_drc.format(mode=mode)
            )


def sound_mode_factory(instance: DenonAVRFoundation) -> DenonAVRSoundMode:
    """Create DenonAVRSoundMode at receiver instances."""
    # pylint: disable=protected-access
    new = DenonAVRSoundMode(device=instance._device)
    return new



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module covers some basic automated tests of Denon AVR receivers.

:copyright: (c) 2016 by Oliver Goetz.
:license: MIT, see LICENSE for more details.
"""

import asyncio
from unittest import mock

import httpx
import pytest
from pytest_httpx import HTTPXMock

import denonavr
from denonavr.api import DenonAVRTelnetApi, DenonAVRTelnetProtocol
from denonavr.const import SOUND_MODE_MAPPING
from denonavr.exceptions import AvrNetworkError, AvrTimoutError

FAKE_IP = "10.0.0.0"

NO_ZONES = None
ZONE2 = {"Zone2": None}
ZONE3 = {"Zone3": None}
ZONE2_ZONE3 = {"Zone2": None, "Zone3": None}

TESTING_RECEIVERS = {
    "AVR-X4100W": (NO_ZONES, denonavr.const.AVR_X),
    "AVR-2312CI": (NO_ZONES, denonavr.const.AVR),
    "AVR-1912": (NO_ZONES, denonavr.const.AVR),
    "AVR-3311CI": (NO_ZONES, denonavr.const.AVR),
    "M-RC610": (NO_ZONES, denonavr.const.AVR_X),
    "AVR-X2100W-2": (NO_ZONES, denonavr.const.AVR_X),
    "AVR-X2000": (ZONE2_ZONE3, denonavr.const.AVR_X),
    "AVR-X2000-2": (NO_ZONES, denonavr.const.AVR_X),
    "SR5008": (NO_ZONES, denonavr.const.AVR_X),
    "M-CR603": (NO_ZONES, denonavr.const.AVR),
    "NR1604": (ZONE2_ZONE3, denonavr.const.AVR_X),
    "AVR-4810": (NO_ZONES, denonavr.const.AVR),
    "AVR-3312": (NO_ZONES, denonavr.const.AVR),
    "NR1609": (ZONE2, denonavr.const.AVR_X_2016),
    "AVC-8500H": (ZONE2_ZONE3, denonavr.const.AVR_X_2016),
    "AVC-A10H": (ZONE2_ZONE3, denonavr.const.AVR_X_2016),
    "AVR-X4300H": (ZONE2_ZONE3, denonavr.const.AVR_X_2016),
    "AVR-X1100W": (ZONE2, denonavr.const.AVR_X),
    "SR6012": (ZONE2, denonavr.const.AVR_X_2016),
    "M-CR510": (NO_ZONES, denonavr.const.AVR_X),
    "M-CR510-2": (NO_ZONES, denonavr.const.AVR_X),
    "AVC-X3700H": (ZONE2, denonavr.const.AVR_X_2016),
    "AVR-X4000": (ZONE2_ZONE3, denonavr.const.AVR_X),
    "SR6011": (ZONE2, denonavr.const.AVR_X),
    "AV7703": (ZONE2_ZONE3, denonavr.const.AVR_X_2016),
    "AVR-1713": (NO_ZONES, denonavr.const.AVR_X),
    "AVR-3313": (ZONE2_ZONE3, denonavr.const.AVR_X),
}

APPCOMMAND_URL = "/goform/AppCommand.xml"
STATUS_URL = "/goform/formMainZone_MainZoneXmlStatus.xml"
STATUS_Z2_URL = "/goform/formZone2_Zone2XmlStatus.xml"
STATUS_Z3_URL = "/goform/formZone3_Zone3XmlStatus.xml"
MAINZONE_URL = "/goform/formMainZone_MainZoneXml.xml"
DEVICEINFO_URL = "/goform/Deviceinfo.xml"
NETAUDIOSTATUS_URL = "/goform/formNetAudio_StatusXml.xml"
TUNERSTATUS_URL = "/goform/formTuner_TunerXml.xml"
HDTUNERSTATUS_URL = "/goform/formTuner_HdXml.xml"
DESCRIPTION_URL1 = "/description.xml"
DESCRIPTION_URL2 = "/upnp/desc/aios_device/aios_device.xml"


def get_sample_content(filename):
    """Return sample content form file."""
    with open(f"tests/xml/{filename}", encoding="utf-8") as file:
        return file.read()


class TestMainFunctions:
    """Test case for main functions of Denon AVR."""

    testing_receiver = None
    denon = None
    future = None

    def custom_matcher(self, request: httpx.Request, *args, **kwargs):
        """Match URLs to sample files."""
        port_suffix = ""

        if request.url.port == 8080:
            port_suffix = "-8080"

        try:
            if request.url.path == STATUS_URL:
                content = get_sample_content(
                    f"{self.testing_receiver}-formMainZone_MainZoneXmlStatus"
                    f"{port_suffix}.xml"
                )
            elif request.url.path == STATUS_Z2_URL:
                content = get_sample_content(
                    f"{self.testing_receiver}-formZone2_Zone2XmlStatus{port_suffix}.xml"
                )
            elif request.url.path == STATUS_Z3_URL:
                content = get_sample_content(
                    f"{self.testing_receiver}-formZone3_Zone3XmlStatus{port_suffix}.xml"
                )
            elif request.url.path == MAINZONE_URL:
                content = get_sample_content(
                    f"{self.testing_receiver}-formMainZone_MainZoneXml{port_suffix}.xml"
                )
            elif request.url.path == DEVICEINFO_URL:
                content = get_sample_content(
                    f"{self.testing_receiver}-Deviceinfo{port_suffix}.xml"
                )
            elif request.url.path == NETAUDIOSTATUS_URL:
                content = get_sample_content(
                    f"{self.testing_receiver}-formNetAudio_StatusXml{port_suffix}.xml"
                )
            elif request.url.path == TUNERSTATUS_URL:
                content = get_sample_content(
                    f"{self.testing_receiver}-formTuner_TunerXml{port_suffix}.xml"
                )
            elif request.url.path == HDTUNERSTATUS_URL:
                content = get_sample_content(
                    f"{self.testing_receiver}-formTuner_HdXml{port_suffix}.xml"
                )
            elif request.url.path == APPCOMMAND_URL:
                content_str = request.read().decode("utf-8")
                if "GetFriendlyName" in content_str:
                    ep_suffix = "-setup"
                elif "GetAllZoneSource" in content_str:
                    ep_suffix = "-update"
                elif "GetSurroundModeStatus" in content_str:
                    ep_suffix = "-update-soundmode"
                else:
                    ep_suffix = "-update-tonecontrol"
                content = get_sample_content(
                    f"{self.testing_receiver}-AppCommand{ep_suffix}{port_suffix}.xml"
                )
            elif request.url.path in [DESCRIPTION_URL1, DESCRIPTION_URL2]:
                content = get_sample_content("AVR-X1600H_upnp.xml")
            else:
                content = "DATA"
        except FileNotFoundError:
            content = "Error 403: Forbidden\nAccess Forbidden"
            status_code = 403
        else:
            status_code = 200

        resp = httpx.Response(status_code=status_code, content=content)

        return resp

    async def _callback(self, zone, event, parameter):
        self.future.set_result(True)

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_receiver_type(self, httpx_mock: HTTPXMock):
        """Check that receiver type is determined correctly."""
        httpx_mock.add_callback(self.custom_matcher)
        for receiver, spec in TESTING_RECEIVERS.items():
            print(f"Receiver: {receiver}")
            # Switch receiver and update to load new sample files
            self.testing_receiver = receiver
            self.denon = denonavr.DenonAVR(FAKE_IP, add_zones=spec[0])
            await self.denon.async_setup()
            assert self.denon.receiver_type == spec[1].type, (
                f"Receiver type is {self.denon.receiver_type} not {spec[1].type} for"
                f" receiver {receiver}"
            )
            assert self.denon.receiver_port == spec[1].port, (
                f"Receiver port is {self.denon.receiver_port} not {spec[1].port} for"
                f" receiver {receiver}"
            )

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_input_func_switch(self, httpx_mock: HTTPXMock):
        """Switch through all input functions of all tested receivers."""
        httpx_mock.add_callback(self.custom_matcher)
        for receiver, spec in TESTING_RECEIVERS.items():
            # Switch receiver and update to load new sample files
            self.testing_receiver = receiver
            self.denon = denonavr.DenonAVR(FAKE_IP, add_zones=spec[0])
            # Switch through all functions and check if successful
            for name, zone in self.denon.zones.items():
                print(f"Receiver: {receiver}, Zone: {name}")
                await self.denon.zones[name].async_update()
                assert len(zone.input_func_list) > 0
                for input_func in zone.input_func_list:
                    await self.denon.zones[name].async_set_input_func(input_func)

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_attributes_not_none(self, httpx_mock: HTTPXMock):
        """Check that certain attributes are not None."""
        httpx_mock.add_callback(self.custom_matcher)
        for receiver, spec in TESTING_RECEIVERS.items():
            print(f"Receiver: {receiver}")
            # Switch receiver and update to load new sample files
            self.testing_receiver = receiver
            self.denon = denonavr.DenonAVR(FAKE_IP, add_zones=spec[0])
            await self.denon.async_setup()
            assert self.denon.name is not None, f"Name is None for receiver {receiver}"
            assert (
                self.denon.support_sound_mode is not None
            ), f"support_sound_mode is None for receiver {receiver}"
            await self.denon.async_update()
            assert (
                self.denon.power is not None
            ), f"Power status is None for receiver {receiver}"
            assert (
                self.denon.state is not None
            ), f"State is None for receiver {receiver}"

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_sound_mode(self, httpx_mock: HTTPXMock):
        """Check if a valid sound mode is returned."""
        httpx_mock.add_callback(self.custom_matcher)
        for receiver, spec in TESTING_RECEIVERS.items():
            # Switch receiver and update to load new sample files
            self.testing_receiver = receiver
            self.denon = denonavr.DenonAVR(FAKE_IP, add_zones=spec[0])
            # Switch through all functions and check if successful
            for name in self.denon.zones:
                print(f"Receiver: {receiver}, Zone: {name}")
                await self.denon.zones[name].async_update()
                support_sound_mode = self.denon.zones[name].support_sound_mode
                sound_mode = self.denon.zones[name].sound_mode
                assert (
                    sound_mode in [*SOUND_MODE_MAPPING, None]
                    or support_sound_mode is not True
                )

    @pytest.mark.asyncio
    async def test_protocol_connected(self):
        """Connected after connection made."""
        proto = DenonAVRTelnetProtocol(None, None)
        transport = mock.Mock(is_closing=lambda: False)
        proto.connection_made(transport)

        assert proto.connected

    @pytest.mark.asyncio
    async def test_protocol_not_connected(self):
        """Not connected when connection not made."""
        proto = DenonAVRTelnetProtocol(None, None)

        assert not proto.connected

    @pytest.mark.asyncio
    async def test_protocol_closing(self):
        """Not connected when connection is closing."""
        proto = DenonAVRTelnetProtocol(None, None)
        transport = mock.Mock(is_closing=lambda: True)
        proto.connection_made(transport)

        assert not proto.connected

    @pytest.mark.asyncio
    async def test_protocol_writes_to_transport(self):
        """Write writes to transport."""
        proto = DenonAVRTelnetProtocol(None, None)
        transport = mock.Mock(is_closing=lambda: False)
        proto.connection_made(transport)

        proto.write("abc")

        transport.write.assert_called_with("abc".encode("utf-8"))

    @pytest.mark.asyncio
    async def test_protocol_close_closes_transport(self):
        """Write writes to transport."""
        proto = DenonAVRTelnetProtocol(None, None)
        transport = mock.Mock()
        proto.connection_made(transport)

        proto.close()

        transport.close.assert_called_once()

    @pytest.mark.asyncio
    async def test_protocol_full_message_triggers_callback(self):
        """Read callback triggered for CR terminated message."""
        data_received_mock = mock.Mock()
        proto = DenonAVRTelnetProtocol(data_received_mock, None)
        transport = mock.Mock(is_closing=lambda: False)
        proto.connection_made(transport)
        proto.data_received(b"abc\r")

        data_received_mock.assert_called_with(b"abc".decode("utf-8"))

    @pytest.mark.asyncio
    async def test_protocol_partial_message_does_not_triggers_callback(self):
        """Read callback not triggered for non CR terminated message."""
        data_received_mock = mock.Mock()
        proto = DenonAVRTelnetProtocol(data_received_mock, None)
        transport = mock.Mock(is_closing=lambda: False)
        proto.connection_made(transport)
        proto.data_received(b"abc")

        data_received_mock.assert_not_called()

    @pytest.mark.asyncio
    async def test_protocol_connection_lost_triggers_callback(self):
        """Connection lost triggers callback."""
        conn_lost_mock = mock.Mock()
        proto = DenonAVRTelnetProtocol(None, conn_lost_mock)
        transport = mock.Mock(is_closing=lambda: False)
        proto.connection_made(transport)
        proto.connection_lost(Exception())

        conn_lost_mock.assert_called_once()

    @pytest.mark.asyncio
    async def test_telnetapi_not_healthy_if_protocol_not_connected(self):
        """Not healthy when not connected."""
        protocol = mock.Mock()
        api = DenonAVRTelnetApi()
        # pylint: disable=protected-access
        api._protocol = protocol
        protocol.connected = False

        assert not api.healthy

    @pytest.mark.asyncio
    async def test_telnetapi_healthy_if_protocol_connected(self):
        """Healthy when connected."""
        protocol = mock.Mock()
        api = DenonAVRTelnetApi()
        # pylint: disable=protected-access
        api._protocol = protocol
        protocol.connected = True

        assert api.healthy

    @pytest.mark.asyncio
    async def test_register_callback_invalid_event_raises_valueerror(self):
        """Callback raises on invalid event."""
        mock_callback = mock.Mock()
        api = DenonAVRTelnetApi()
        with pytest.raises(ValueError):
            api.register_callback("INVALIDEVENT", mock_callback.method)

    @pytest.mark.asyncio
    async def test_register_callback_valid_event_does_not_raise(self):
        """Callback succeeds on valid event."""
        mock_callback = mock.Mock()
        api = DenonAVRTelnetApi()
        api.register_callback("ALL", mock_callback.method)

    @pytest.mark.asyncio
    async def test_unregister_callback_succeeds(self):
        """Unregister callback succeeds."""
        mock_callback = mock.Mock()
        api = DenonAVRTelnetApi()
        api.unregister_callback("ALL", mock_callback.method)

    @pytest.mark.asyncio
    async def test_connect_connectionrefused_raises_networkerror(self):
        """Connect raises NetworkError when ConnectionRefused."""
        api = DenonAVRTelnetApi()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=ConnectionRefusedError()
            )
            with pytest.raises(AvrNetworkError):
                await api.async_connect()

    @pytest.mark.asyncio
    async def test_connect_oserror_raises_networkerror(self):
        """Connect raises NetworkError when OSError."""
        api = DenonAVRTelnetApi()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=OSError()
            )
            with pytest.raises(AvrNetworkError):
                await api.async_connect()

    @pytest.mark.asyncio
    async def test_connect_ioerror_raises_networkerror(self):
        """Connect raises NetworkError when IOError."""
        api = DenonAVRTelnetApi()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=IOError()
            )
            with pytest.raises(AvrNetworkError):
                await api.async_connect()

    @pytest.mark.asyncio
    async def test_connect_timeouterror_raises_timeouterror(self):
        """Connect raises AvrTimeoutError when TimeoutError."""
        api = DenonAVRTelnetApi()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=asyncio.TimeoutError()
            )
            with pytest.raises(AvrTimoutError):
                await api.async_connect()

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_receive_callback_called(self, httpx_mock: HTTPXMock):
        """Check that the callback is triggered whena message is received."""
        transport = mock.Mock(is_closing=lambda: False)
        protocol = DenonAVRTelnetProtocol(None, None)

        def create_conn(proto_lambda, host, port):
            proto = proto_lambda()
            # pylint: disable=protected-access
            protocol._on_connection_lost = proto._on_connection_lost
            # pylint: disable=protected-access
            protocol._on_message = proto._on_message
            proto.connection_made(transport)
            protocol.connection_made(transport)
            return [transport, proto]

        httpx_mock.add_callback(self.custom_matcher)

        self.denon = denonavr.DenonAVR(FAKE_IP)
        # pylint: disable=protected-access
        self.denon._device.telnet_api._send_confirmation_timeout = 0.1
        await self.denon.async_setup()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=create_conn
            )
            await self.denon.async_telnet_connect()
            mock_obj = mock.Mock()
            self.future = asyncio.Future()
            self.denon.register_callback("ALL", mock_obj.method)
            self.denon.register_callback("ALL", self._callback)
            protocol.data_received(b"MUON\r")
            await self.future
            mock_obj.method.assert_called_once()

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_mute_on(self, httpx_mock: HTTPXMock):
        """Check that mute on is processed."""
        transport = mock.Mock(is_closing=lambda: False)
        protocol = DenonAVRTelnetProtocol(None, None)

        def create_conn(proto_lambda, host, port):
            proto = proto_lambda()
            # pylint: disable=protected-access
            protocol._on_message = proto._on_message
            proto.connection_made(transport)
            protocol.connection_made(transport)
            return [transport, proto]

        httpx_mock.add_callback(self.custom_matcher)
        self.denon = denonavr.DenonAVR(FAKE_IP)
        # pylint: disable=protected-access
        self.denon._device.telnet_api._send_confirmation_timeout = 0.1
        await self.denon.async_setup()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=create_conn
            )
            await self.denon.async_telnet_connect()
            self.future = asyncio.Future()
            self.denon.register_callback("ALL", self._callback)
            protocol.data_received(b"MUON\r")
            await self.future
            assert self.denon.muted

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_mute_off(self, httpx_mock: HTTPXMock):
        """Check that mute off is processed."""
        transport = mock.Mock(is_closing=lambda: False)
        protocol = DenonAVRTelnetProtocol(None, None)

        def create_conn(proto_lambda, host, port):
            proto = proto_lambda()
            # pylint: disable=protected-access
            protocol._on_message = proto._on_message
            proto.connection_made(transport)
            protocol.connection_made(transport)
            return [transport, proto]

        httpx_mock.add_callback(self.custom_matcher)
        self.denon = denonavr.DenonAVR(FAKE_IP)
        # pylint: disable=protected-access
        self.denon._device.telnet_api._send_confirmation_timeout = 0.1
        await self.denon.async_setup()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=create_conn
            )
            await self.denon.async_telnet_connect()
            self.future = asyncio.Future()
            self.denon.register_callback("ALL", self._callback)
            protocol.data_received(b"MUOFF\r")
            await self.future
            assert not self.denon.muted

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_power_on(self, httpx_mock: HTTPXMock):
        """Check that power on is processed."""
        transport = mock.Mock(is_closing=lambda: False)
        protocol = DenonAVRTelnetProtocol(None, None)

        def create_conn(proto_lambda, host, port):
            proto = proto_lambda()
            # pylint: disable=protected-access
            protocol._on_message = proto._on_message
            proto.connection_made(transport)
            protocol.connection_made(transport)
            return [transport, proto]

        httpx_mock.add_callback(self.custom_matcher)
        self.denon = denonavr.DenonAVR(FAKE_IP)
        # pylint: disable=protected-access
        self.denon._device.telnet_api._send_confirmation_timeout = 0.1
        await self.denon.async_setup()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=create_conn
            )
            await self.denon.async_telnet_connect()
            self.future = asyncio.Future()
            self.denon.register_callback("ALL", self._callback)
            protocol.data_received(b"ZMON\r")
            await self.future
            assert self.denon.power == "ON"

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_power_off(self, httpx_mock: HTTPXMock):
        """Check that power off is processed."""
        transport = mock.Mock(is_closing=lambda: False)
        protocol = DenonAVRTelnetProtocol(None, None)

        def create_conn(proto_lambda, host, port):
            proto = proto_lambda()
            # pylint: disable=protected-access
            protocol._on_message = proto._on_message
            proto.connection_made(transport)
            protocol.connection_made(transport)
            return [transport, proto]

        httpx_mock.add_callback(self.custom_matcher)
        self.denon = denonavr.DenonAVR(FAKE_IP)
        # pylint: disable=protected-access
        self.denon._device.telnet_api._send_confirmation_timeout = 0.1
        await self.denon.async_setup()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=create_conn
            )
            await self.denon.async_telnet_connect()
            self.future = asyncio.Future()
            self.denon.register_callback("ALL", self._callback)
            protocol.data_received(b"ZMOFF\r")
            await self.future
            assert self.denon.power == "OFF"

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_volume_min(self, httpx_mock: HTTPXMock):
        """Check that minimum volume is processed."""
        transport = mock.Mock(is_closing=lambda: False)
        protocol = DenonAVRTelnetProtocol(None, None)

        def create_conn(proto_lambda, host, port):
            proto = proto_lambda()
            # pylint: disable=protected-access
            protocol._on_message = proto._on_message
            proto.connection_made(transport)
            protocol.connection_made(transport)
            return [transport, proto]

        httpx_mock.add_callback(self.custom_matcher)
        self.denon = denonavr.DenonAVR(FAKE_IP)
        # pylint: disable=protected-access
        self.denon._device.telnet_api._send_confirmation_timeout = 0.1
        await self.denon.async_setup()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=create_conn
            )
            await self.denon.async_telnet_connect()
            self.future = asyncio.Future()
            self.denon.register_callback("ALL", self._callback)
            protocol.data_received(b"MV00\r")
            await self.future
            assert self.denon.volume == -80.0

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_volume_wholenumber(self, httpx_mock: HTTPXMock):
        """Check that whole number volume is processed."""
        transport = mock.Mock(is_closing=lambda: False)
        protocol = DenonAVRTelnetProtocol(None, None)

        def create_conn(proto_lambda, host, port):
            proto = proto_lambda()
            # pylint: disable=protected-access
            protocol._on_message = proto._on_message
            proto.connection_made(transport)
            protocol.connection_made(transport)
            return [transport, proto]

        httpx_mock.add_callback(self.custom_matcher)
        self.denon = denonavr.DenonAVR(FAKE_IP)
        # pylint: disable=protected-access
        self.denon._device.telnet_api._send_confirmation_timeout = 0.1
        await self.denon.async_setup()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=create_conn
            )
            await self.denon.async_telnet_connect()
            self.future = asyncio.Future()
            self.denon.register_callback("ALL", self._callback)
            protocol.data_received(b"MV56\r")
            await self.future
            assert self.denon.volume == -24.0

    @pytest.mark.asyncio
    @pytest.mark.httpx_mock(can_send_already_matched_responses=True)
    async def test_volume_fraction(self, httpx_mock: HTTPXMock):
        """Check that fractional volume is processed."""
        transport = mock.Mock(is_closing=lambda: False)
        protocol = DenonAVRTelnetProtocol(None, None)

        def create_conn(proto_lambda, host, port):
            proto = proto_lambda()
            # pylint: disable=protected-access
            protocol._on_message = proto._on_message
            proto.connection_made(transport)
            protocol.connection_made(transport)
            return [transport, proto]

        httpx_mock.add_callback(self.custom_matcher)
        self.denon = denonavr.DenonAVR(FAKE_IP)
        # pylint: disable=protected-access
        self.denon._device.telnet_api._send_confirmation_timeout = 0.1
        await self.denon.async_setup()
        with mock.patch("asyncio.get_event_loop", new_callable=mock.Mock) as debug_mock:
            debug_mock.return_value.create_connection = mock.AsyncMock(
                side_effect=create_conn
            )
            await self.denon.async_telnet_connect()
            self.future = asyncio.Future()
            self.denon.register_callback("ALL", self._callback)
            protocol.data_received(b"MV565\r")
            await self.future
            assert self.denon.volume == -23.5
